"""
Code generation for C preprocessing directives.
"""

from ctree.codegen import CodeGenVisitor


class CppCodeGen(CodeGenVisitor):
    """
    Visitor to generate C preprocessor directives.
    """

    def visit_CppInclude(self, node):
        if node.angled_brackets:
            return "#include <%s>" % node.target
        else:
            return '#include "%s"' % node.target

    def visit_CppComment(self, node):
        return "// " + ("\n" + self._tab() + "// ").join(
            node.text.splitlines())

    def visit_CppDefine(self, node):
        params = ", ".join(map(str, node.params))
        return "#define %s(%s) (%s)" % (node.name, params, node.body)



import unittest
from textwrap import dedent

from ctree.c.nodes import *
from ctree.analyses import *
from ctree.util import flatten, enumerate_flatten


a = SymbolRef("a")
b = SymbolRef("b")
c = SymbolRef("c")

class TestRawFlattening(unittest.TestCase):
    def _check(self, nested, flat):
        actual, expected = list(flatten(nested)), flat
        self.assertListEqual(actual, expected)

    def test_obj(self):
        self._check(123, [123])

    def test_flat(self):
        self._check([1, 2, 3], [1, 2, 3])

    def test_lol(self):
        self._check([1, [2, 3]],   [1, 2, 3])
        self._check([[1, 2], 3],   [1, 2, 3])
        self._check([[1], 2, [3]], [1, 2, 3])

    def test_lolol(self):
        self._check([1, [2, [3, 4]]], [1, 2, 3, 4])
        self._check([[1, [2]], 3, 4], [1, 2, 3, 4])


class TestRawEnumeratedFlattening(unittest.TestCase):
    def _check(self, nested, exp_indices):
        act_indices, values = zip(*enumerate_flatten(nested))
        for actual, expected in zip(act_indices, exp_indices):
            self.assertTupleEqual(actual, expected)

    def test_obj(self):
        self._check(123, ())

    def test_flat(self):
        self._check([1, 2, 3], [(0,), (1,), (2,)])

    def test_lol(self):
        self._check([1, [2, 3]],   [(0,), (1,0), (1,1)])
        self._check([[1, 2], 3],   [(0,0), (0,1), (1,)])
        self._check([[1], 2, [3]], [(0,0), (1,), (2,0)])

    def test_lolol(self):
        self._check([1, [2, [3, 4]]], [(0,), (1,0), (1,1,0), (1,1,1)])
        self._check([[1, [2]], 3, 4], [(0,0), (0,1,0), (1,), (2,)])


class TestListFlattening(unittest.TestCase):
    def _check(self, tree, expected):
        actual = str(tree)
        self.assertEqual(actual, dedent(expected))

    def test_flat_1(self):
        tree = Block([a])
        self._check(tree, """\
        {
            a;
        }""")

    def test_flat_3(self):
        tree = Block([a, b, c])
        self._check(tree, """\
        {
            a;
            b;
            c;
        }""")

    def test_lol_1(self):
        tree = Block([[a]])
        self._check(tree, """\
        {
            a;
        }""")

    def test_lol_3(self):
        tree = Block([[a, b], c])
        self._check(tree, """\
        {
            a;
            b;
            c;
        }""")

    def test_lolol_3(self):
        tree = Block([[a, b], [[[c]]]])
        self._check(tree, """\
        {
            a;
            b;
            c;
        }""")


@unittest.skip("fails")
class TestFlatteningDotGen(unittest.TestCase):
    def test_lol_1(self):
        tree = Block([[a, b]])
        tree.to_dot()



from __future__ import print_function

"""
base class for generating code appropriate to the selected backend
"""
from ctree.visitors import NodeVisitor
from ctree.util import flatten


class CodeGenVisitor(NodeVisitor):
    """
    Return a string containing the program text.
    """

    def __init__(self, indent=0):
        self._indent = indent

    # -------------------------------------------------------------------------
    # common support methods

    def _tab(self):
        """return correct spaces if tab found"""
        return "    " * self._indent

    def _genblock(self, forest, insert_curly_brackets=True,
                  increase_indent=True):
        """generate block of code adding semi colons as necessary"""
        if increase_indent:
            self._indent += 1
        body = ""
        for tree in flatten(forest):
            semicolon_opt = ";" if tree._requires_semicolon() else ""
            block = tree.codegen(self._indent)
            if block is not "":
                body += self._tab() + block + semicolon_opt + "\n"
        if increase_indent:
            self._indent -= 1
        if insert_curly_brackets:
            return "{\n%s%s}" % (body, self._tab())
        else:
            return "\n%s" % body

    def _parenthesize(self, parent, child):
        """A format string that includes parentheses if needed."""
        if self._requires_parentheses(parent, child) or \
                child._force_parentheses is True:
            return "(%s)" % child
        else:
            return "%s" % child

    def _requires_parentheses(self, parent, child):
        """True by default."""
        return True



import unittest

from ctree.util import highlight
from util import PreventImport

class TestNoPygments(unittest.TestCase):
    def test_no_pygments(self):
        """ No pygments => no syntax highlighting. """
        with PreventImport('pygments'):
            code = "int a = 0;"
            highlighted = highlight(code, "c")
            self.assertEqual(code, highlighted)


class TestHighlights(unittest.TestCase):
    def test_highlight_c(self):
        highlighted = highlight("int a = 0;", "c")
        assert len(highlighted) > 0

    def test_highlight_llvm(self):
        highlighted = highlight("%add.i.1 = fadd double %14, %mul.i.1", "llvm")
        assert len(highlighted) > 0

    def test_bad_lang(self):
        with self.assertRaises(ValueError):
            highlight("int a = 0;", "not-a-language")



import unittest
import ctypes as ct

from ctree.c.nodes import *


class TestSymbols(unittest.TestCase):
    def _check(self, actual, expected):
        self.assertEqual(actual.codegen(), expected)

    def test_symbolref(self):
        ref = SymbolRef("foo")
        self._check(ref, "foo")

    def test_init_local(self):
        ref = SymbolRef("foo", _local=True)
        self._check(ref, "__local foo")

    def test_init_const(self):
        ref = SymbolRef("foo", _const=True)
        self._check(ref, "const foo")

    def test_set_local(self):
        ref = SymbolRef("foo")
        ref.set_local()
        self._check(ref, "__local foo")

    def test_set_const(self):
        ref = SymbolRef("foo")
        ref.set_const()
        self._check(ref, "const foo")

    def test_set_global(self):
        ref = SymbolRef("foo")
        ref.set_global()
        self._check(ref, "__global foo")

    def test_unique(self):
        ref1 = SymbolRef.unique("foo", ct.c_int())
        ref2 = SymbolRef.unique("foo", ct.c_int())
        self.assertNotEqual(ref1.codegen(), ref2.codegen())

    def test_copy(self):
        ref1 = SymbolRef("foo")
        ref2 = ref1.copy()
        self._check(ref1, ref2.codegen(()))

    def test_copy_without_declare(self):
        ref1 = SymbolRef("foo", ct.c_int())
        ref2 = ref1.copy()
        self._check(ref2, "foo")

    def test_copy_with_declare(self):
        ref1 = SymbolRef("foo", ct.c_int())
        ref2 = ref1.copy(declare=True)
        self._check(ref2, "int foo")




"""helper for getting properly formatted ast from arbitrary python object"""
import ast
import inspect
import textwrap


def get_ast(obj):
    """
    Return the Python ast for the given object, which may
    be anything that inspect.getsource accepts (incl.
    a module, class, method, function, traceback, frame,
    or code object).
    """
    indented_program_txt = inspect.getsource(obj)
    program_txt = textwrap.dedent(indented_program_txt)
    return ast.parse(program_txt)


"""
A pretty-printing dump function for the ast module.  The code was copied from
the ast.dump function and modified slightly to pretty-print.

Alex Leone (acleone ~AT~ gmail.com), 2010-01-30

From http://alexleone.blogspot.co.uk/2010/01/python-ast-pretty-printer.html
"""

from ast import *

def dump(node, annotate_fields=True, include_attributes=False, indent='  '):
    """
    Return a formatted dump of the tree in *node*.  This is mainly useful for
    debugging purposes.  The returned string will show the names and the values
    for fields.  This makes the code impossible to evaluate, so if evaluation is
    wanted *annotate_fields* must be set to False.  Attributes such as line
    numbers and column offsets are not dumped by default.  If this is wanted,
    *include_attributes* can be set to True.
    """
    def _format(node, level=0):
        if isinstance(node, AST):
            fields = [(a, _format(b, level)) for a, b in iter_fields(node)]
            if include_attributes and node._attributes:
                fields.extend([(a, _format(getattr(node, a), level))
                               for a in node._attributes])
            return ''.join([
                node.__class__.__name__,
                '(',
                ', '.join(('%s=%s' % field for field in fields)
                           if annotate_fields else
                           (b for a, b in fields)),
                ')'])
        elif isinstance(node, list):
            lines = ['[']
            lines.extend((indent * (level + 2) + _format(x, level + 2) + ','
                         for x in node))
            if len(lines) > 1:
                lines.append(indent * (level + 1) + ']')
            else:
                lines[-1] += ']'
            return '\n'.join(lines)
        return repr(node)

    if not isinstance(node, AST):
        raise TypeError('expected AST, got %r' % node.__class__.__name__)
    return _format(node)

def parseprint(code, filename="<string>", mode="exec", **kwargs):
    """Parse some code from a string and pretty-print it."""
    node = parse(code, mode=mode)   # An ode to the code
    print(dump(node, **kwargs))

# Short name: pdp = parse, dump, print
pdp = parseprint

def load_ipython_extension(ip):  # pragma: no cover
    from IPython.core.magic import Magics, magics_class, cell_magic
    from IPython.core import magic_arguments

    @magics_class
    class AstMagics(Magics):

        @magic_arguments.magic_arguments()
        @magic_arguments.argument(
            '-m', '--mode', default='exec',
            help="The mode in which to parse the code. Can be exec (the default), "
                 "eval or single."
        )
        @cell_magic
        def dump_ast(self, line, cell):
            """Parse the code in the cell, and pretty-print the AST."""
            args = magic_arguments.parse_argstring(self.dump_ast, line)
            parseprint(cell, mode=args.mode)

    ip.register_magics(AstMagics)

if __name__ == '__main__':  # pragma: no cover
    import sys, tokenize
    for filename in sys.argv[1:]:
        print('=' * 50)
        print('AST tree for', filename)
        print('=' * 50)
        with tokenize.open(filename) as f:
            fstr = f.read()

        parseprint(fstr, filename=filename, include_attributes=True)
        print()



"""
C preprocessor nodes supported by ctree.
"""

from ctree.nodes import CtreeNode


class CppNode(CtreeNode):
    """Base class for all C Preprocessor nodes in ctree."""

    def codegen(self, indent=0):
        from ctree.cpp.codegen import CppCodeGen

        return CppCodeGen(indent).visit(self)

    def label(self):
        from ctree.cpp.dotgen import CppDotLabeller

        return CppDotLabeller().visit(self)

    def _requires_semicolon(self):
        return False


class CppInclude(CppNode):
    """Represents #include <foo.h>."""

    def __init__(self, target="", angled_brackets=True):
        self.target = target
        self.angled_brackets = angled_brackets


class CppComment(CppNode):
    """Represents // foo"""

    def __init__(self, text=""):
        self.text = text


class CppDefine(CppNode):
    _fields = ['name', 'params', 'body']

    def __init__(self, name=None, params=None, body=None):
        self.name = name
        self.params = params
        self.body = body
        super(CppDefine, self).__init__()



"""
DOT generation for OpenCL.
"""

from ctree.dotgen import DotGenLabeller


class OclDotLabeller(DotGenLabeller):
    """
    Visitor to generator DOT.
    """
    def visit_OclFile(self, node):
        return node.get_filename()



"""
OpenCL nodes supported by ctree.
"""

from ctree.nodes import *
import hashlib


class OclNode(CtreeNode):
    """Base class for all OpenCL nodes supported by ctree."""

    def codegen(self, indent=0):
        """generate ocl code for this node"""
        from ctree.ocl.codegen import OclCodeGen

        return OclCodeGen(indent).visit(self)

    def label(self, indent=0):
        """generate dot element for this node"""
        from ctree.ocl.dotgen import OclDotLabeller

        return OclDotLabeller().visit(self)


class OclFile(OclNode, File):
    """Represents a .cl file."""
    _ext = "cl"

    def __init__(self, name="generated", body=None, path = None):
        #TODO: Inspect complains about 2 args to __init__
        super(OclFile, self).__init__(name, body, path)

    def _compile(self, program_text):
        """
        write the ocl program to a text file and compile it
        """
        import os
        new_hash = hashlib.sha512(program_text.strip().encode()).hexdigest()
        recreate_source = program_text != self._empty and new_hash != self.program_hash
        self.program_hash = new_hash
        cl_src_file = os.path.join(self.path, self.get_filename())
        if recreate_source:
            log.info('Recreating source')
            log.info("file for generated OpenCL: %s" % cl_src_file)
            log.info("generated OpenCL code: (((\n%s\n)))" % program_text)

            # write program text to CL file
            with open(cl_src_file, 'w') as cl_file:
                cl_file.write(program_text)
        else:
            log.info("OpenCL file already generated")
        return None

    def codegen(self, indent=0):
        if self.body:
            return super(OclFile, self).codegen(indent)
        with open(os.path.join(self.path, self.get_filename())) as cl_file:
            return cl_file.read()



import numpy as np

from ctree.types import (
    codegen_type,
    register_type_recognizers,
    register_type_codegenerators,
)

def codegen_ndptr(ndptr):
    prefix = ""
    if getattr(ndptr, "_global", False):
        prefix += "__global "
    return prefix + "%s*" % codegen_type(ndptr._dtype_.type())

register_type_recognizers({
    np.ndarray: lambda obj: np.ctypeslib.as_ctypes(obj)
})

register_type_codegenerators({
    # pointers
    np.ctypeslib._ndptr: codegen_ndptr,

    # boolean types
    np.bool8:            lambda t: "bool",

    # signed integer types
    np.byte:             lambda t: "char",
    np.short:            lambda t: "short",
    np.intc:             lambda t: "int",
    np.longlong:         lambda t: "long long",

    # technically not universally compatible
    np.int8:             lambda t: "char",
    np.int16:            lambda t: "short",
    np.int32:            lambda t: "int",
    np.int64:            lambda t: "long",

    # unsigned integer types
    np.ubyte:            lambda t: "unsigned byte",
    np.ushort:           lambda t: "unsigned short",
    np.uintc:            lambda t: "unsigned int",
    np.ulonglong:        lambda t: "unsigned long long",

    # floating point types
    np.single:           lambda t: "float",
    np.float32:          lambda t: "float",
    np.double:           lambda t: "double",
    np.float64:          lambda t: "double",
})



import ctypes

from ctree.types import get_ctype, get_common_ctype
from util import CtreeTest
from ctree.c.nodes import SymbolRef, FunctionDecl

class TestTypeRecognizer(CtreeTest):
    def test_int(self):
        ty = get_ctype(123)
        self.assertIsInstance(ty, ctypes.c_long)

    def test_float(self):
        ty = get_ctype(456.7)
        self.assertIsInstance(ty, ctypes.c_double)

    def test_char(self):
        ty = get_ctype("c")
        self.assertIsInstance(ty, ctypes.c_char)

    def test_none(self):
        ty = get_ctype(None)
        self.assertIsInstance(ty, type(None))

    def test_bool(self):
        ty = get_ctype(True)
        self.assertIsInstance(ty, ctypes.c_bool)

    def test_string(self):
        self.assertIsInstance(get_ctype("foo"),     ctypes.c_char_p)
        self.assertIsInstance(get_ctype(""),        ctypes.c_char_p)
        self.assertIsInstance(get_ctype("one two"), ctypes.c_char_p)

    def test_bad_type(self):
        class Bad(object): pass
        with self.assertRaises(ValueError):
            ty = get_ctype(Bad())


class TestTypeCodeGen(CtreeTest):
    def test_int(self):
        tree = SymbolRef("i", ctypes.c_int())
        self._check_code(tree, "int i")

    def test_long(self):
        tree = SymbolRef("i", ctypes.c_long())
        self._check_code(tree, "long i")

    def test_float(self):
        tree = SymbolRef("i", ctypes.c_double())
        self._check_code(tree, "double i")

    def test_char(self):
        tree = SymbolRef("i", ctypes.c_char())
        self._check_code(tree, "char i")

    def test_none(self):
        tree = SymbolRef("i", ctypes.c_void_p())
        self._check_code(tree, "void* i")

    def test_bool(self):
        tree = SymbolRef("i", ctypes.c_bool())
        self._check_code(tree, "bool i")

    def test_string(self):
        tree = SymbolRef("i", ctypes.c_char_p())
        self._check_code(tree, "char* i")

    def test_pointer(self):
        tree = SymbolRef("i", ctypes.POINTER(ctypes.c_double)())
        self._check_code(tree, "double* i")

    def test_none(self):
        tree = FunctionDecl(None, "foo", ())
        self._check_code(tree, "void foo()")

    def test_bad_type(self):
        class Bad(object): pass
        with self.assertRaises(ValueError):
            SymbolRef("i", Bad()).codegen()

class TestTypeCoercion(CtreeTest):
    def test_coercion(self):
        types = (ctypes.c_long, ctypes.c_double, ctypes.c_int)
        self.assertEqual(get_common_ctype(types), ctypes.c_double)



"""
Macros for simplifying construction of OpenCL
programs.
"""

import ast
from ctypes import c_size_t

from ctree.c.nodes import SymbolRef, Block, Assign, FunctionCall
from ctree.c.nodes import If, Eq, NotEq, Or, Not, Ref, Constant, String
from ctree.c.macros import NULL, printf


def CL_DEVICE_TYPE_GPU():
    return SymbolRef("CL_DEVICE_TYPE_GPU")


def CL_DEVICE_TYPE_CPU():
    return SymbolRef("CL_DEVICE_TYPE_CPU")


def CL_DEVICE_TYPE_ACCELERATOR():
    return SymbolRef("CL_DEVICE_TYPE_ACCELERATOR")


def CL_DEVICE_TYPE_DEFAULT():
    return SymbolRef("CL_DEVICE_TYPE_DEFAULT")


def CL_DEVICE_TYPE_ALL():
    return SymbolRef("CL_DEVICE_TYPE_ALL")


def CL_SUCCESS():
    return SymbolRef("CL_SUCCESS")


def CLK_LOCAL_MEM_FENCE():
    return SymbolRef("CLK_LOCAL_MEM_FENCE")

def CLK_GLOBAL_MEM_FENCE():
    return SymbolRef("CLK_GLOBAL_MEM_FENCE")


def barrier(arg):
    return FunctionCall(SymbolRef('barrier'), [arg])


def get_local_id(id):
    return FunctionCall(SymbolRef('get_local_id'), [Constant(id)])


def get_global_id(id):
    return FunctionCall(SymbolRef('get_global_id'), [Constant(id)])


def get_group_id(id):
    return FunctionCall(SymbolRef('get_group_id'), [Constant(id)])


def get_local_size(id):
    return FunctionCall(SymbolRef('get_local_size'), [Constant(id)])


def get_global_size(id):
    return FunctionCall(SymbolRef('get_global_size'), [Constant(id)])


def get_num_groups(id):
    return FunctionCall(SymbolRef('get_num_groups'), [Constant(id)])


def clReleaseMemObject(arg):
    return FunctionCall(SymbolRef('clReleaseMemObject'), [arg])


def clEnqueueWriteBuffer(queue, buf, blocking, offset, cb, ptr, num_events=0, evt_list_ptr=None, evt=None):
    if     isinstance(buf, str):              buf = SymbolRef(buf)
    if     isinstance(blocking, bool):        blocking = Constant(int(blocking))
    if     isinstance(ptr, str):              ptr = SymbolRef(ptr)
    if not isinstance(offset, ast.AST):       offset = Constant(offset)
    if not isinstance(cb, ast.AST):           cb = Constant(cb)
    if not isinstance(num_events, ast.AST):   num_events = Constant(num_events)
    if not isinstance(evt_list_ptr, ast.AST): event_list_ptr = NULL()
    if not isinstance(evt, ast.AST):          evt = NULL()
    return FunctionCall(SymbolRef('clEnqueueWriteBuffer'), [
        queue, buf, blocking, offset, cb, ptr, num_events, event_list_ptr, evt])

def clEnqueueReadBuffer(queue, buf, blocking, offset, cb, ptr, num_events=0, evt_list_ptr=None, evt=None):
    if     isinstance(buf, str):              buf = SymbolRef(buf)
    if     isinstance(blocking, bool):        blocking = Constant(int(blocking))
    if     isinstance(ptr, str):              ptr = SymbolRef(ptr)
    if not isinstance(offset, ast.AST):       offset = Constant(offset)
    if not isinstance(cb, ast.AST):           cb = Constant(cb)
    if not isinstance(num_events, ast.AST):   num_events = Constant(num_events)
    if not isinstance(evt_list_ptr, ast.AST): event_list_ptr = NULL()
    if not isinstance(evt, ast.AST):          evt = NULL()
    return FunctionCall(SymbolRef('clEnqueueReadBuffer'), [
        queue, buf, blocking, offset, cb, ptr, num_events, event_list_ptr, evt])

def clEnqueueCopyBuffer(queue, src_buf, dst_buf, src_offset=0, dst_offset=0, cb=0):
    if isinstance(src_buf, str):      src_buf = SymbolRef(src_buf)
    if isinstance(dst_buf, str):      dst_buf = SymbolRef(dst_buf)
    if isinstance(src_offset, int):   src_offset = Constant(src_offset)
    if isinstance(dst_offset, int):   dst_offset = Constant(dst_offset)
    if isinstance(cb, int):           cb = Constant(cb)

    num_events = Constant(0)
    event_list_ptr = NULL()
    evt = NULL()

    return FunctionCall(SymbolRef('clEnqueueCopyBuffer'), [
        queue, src_buf, dst_buf, src_offset, dst_offset, cb, num_events, event_list_ptr, evt])

def clSetKernelArg(kernel, arg_index, arg_size, arg_value):
    if isinstance(kernel, str): kernel = SymbolRef(kernel)
    if isinstance(arg_index, int): arg_index = Constant(arg_index)
    if isinstance(arg_size, int): arg_size = Constant(arg_size)
    if isinstance(arg_value, str): arg_value = Ref(SymbolRef(arg_value))
    return FunctionCall(SymbolRef("clSetKernelArg"),
        [kernel, arg_index, arg_size, arg_value])

def clEnqueueNDRangeKernel(queue, kernel, work_dim=1, work_offset=0, global_size=0, local_size=0):
    assert isinstance(queue, SymbolRef)
    assert isinstance(kernel, SymbolRef)
    global_size_sym = SymbolRef('global_size', c_size_t())
    local_size_sym = SymbolRef('local_size', c_size_t())
    call = FunctionCall(SymbolRef("clEnqueueNDRangeKernel"), [
        queue, kernel,
        work_dim, work_offset,
        Ref(global_size_sym.copy()), Ref(local_size_sym.copy()),
        0, NULL(), NULL()
    ])

    return Block([
        Assign(global_size_sym, Constant(global_size)),
        Assign(local_size_sym, Constant(local_size)),
        call
    ])



import ast

from ctree.visitors import NodeVisitor
from ctree.util import enumerate_flatten


def label_for_py_ast_nodes(self):
    from ctree.py.dotgen import PyDotLabeller

    return PyDotLabeller().visit(self)

def to_dot_outer_for_py_ast_nodes(self):
    return "digraph mytree {\n%s}" % self._to_dot()

def to_dot_inner_for_py_ast_nodes(self):
    from ctree.dotgen import DotGenVisitor

    return DotGenVisitor().visit(self)

"""
Bind to_dot_for_py_ast_nodes to all classes that derive from ast.AST. Ideally
we'd be able to bind one method to ast.AST, but it's a built-in type so we
can't.
"""
for entry in ast.__dict__.values():
    try:
        if issubclass(entry, ast.AST):
            entry.label   = label_for_py_ast_nodes
            entry.to_dot  = to_dot_outer_for_py_ast_nodes
            entry._to_dot = to_dot_inner_for_py_ast_nodes
    except TypeError:
        pass


class DotGenLabeller(NodeVisitor):
    def generic_visit(self, node):
        return ""


class DotGenVisitor(NodeVisitor):
    """
    Generates a representation of the AST in the DOT graph language.
    See http://en.wikipedia.org/wiki/DOT_(graph_description_language)
    """
    @staticmethod
    def _qualified_name(obj):
        """
        return object name with leading module
        """
        return "%s.%s" % (obj.__module__, obj.__name__)

    def label(self, node):
        """
        A string to provide useful information for visualization, debugging, etc.
        """
        return r"%s\n%s" % (type(node).__name__, node.label())

    def generic_visit(self, node):

        # label this node
        out_string = 'n%s [label="%s"];\n' % (id(node), self.label(node))

        # edges to children
        for fieldname, fieldvalue in ast.iter_fields(node):
            for index, child in enumerate_flatten(fieldvalue):
                if isinstance(child, ast.AST):
                    suffix = "".join(["[%d]" % i for i in index])
                    out_string += 'n{} -> n{} [label="{}{}"];\n'.format(
                        id(node), id(child), fieldname, suffix)
                    out_string += self.visit(child)
        return out_string



import unittest

from ctree.c.nodes import *


class TestBinaryOps(unittest.TestCase):
    def setUp(self):
        self.foo, self.bar = SymbolRef('foo'), SymbolRef('bar')

    def _check(self, op_cls, expected_string):
        node = op_cls(self.foo, self.bar)
        self.assertEqual(str(node), expected_string)

    def test_add(self):
        self._check(Add, "foo + bar")

    def test_sub(self):
        self._check(Sub, "foo - bar")

    def test_mul(self):
        self._check(Mul, "foo * bar")

    def test_div(self):
        self._check(Div, "foo / bar")

    def test_mod(self):
        self._check(Mod, "foo % bar")

    def test_gt(self):
        self._check(Gt, "foo > bar")

    def test_gte(self):
        self._check(GtE, "foo >= bar")

    def test_lt(self):
        self._check(Lt, "foo < bar")

    def test_lte(self):
        self._check(LtE, "foo <= bar")

    def test_eq(self):
        self._check(Eq, "foo == bar")

    def test_neq(self):
        self._check(NotEq, "foo != bar")

    def test_and(self):
        self._check(And, "foo && bar")

    def test_or(self):
        self._check(Or, "foo || bar")

    def test_bitxor(self):
        self._check(BitXor, "foo ^ bar")

    def test_bitand(self):
        self._check(BitAnd, "foo & bar")

    def test_bitor(self):
        self._check(BitOr, "foo | bar")

    def test_bitshl(self):
        self._check(BitShL, "foo << bar")

    def test_bitshr(self):
        self._check(BitShR, "foo >> bar")

    def test_comma(self):
        self._check(Comma, "foo , bar")

    def test_dot(self):
        self._check(Dot, "foo . bar")

    def test_arrow(self):
        self._check(Arrow, "foo -> bar")


class TestArrayRefs(unittest.TestCase):
    def _check(self, tree, expected):
        actual = str(tree)
        self.assertEqual(actual, expected)

    def test_arrayref_00(self):
        foo, bar = SymbolRef('foo'), SymbolRef('bar')
        tree = ArrayRef(foo, bar)
        self._check(tree, "foo[bar]")



import unittest

from ctree.c.nodes import *


class TestFunctionCalls(unittest.TestCase):
    def _check(self, tree, expected):
        actual = str(tree)
        self.assertEqual(actual, expected)

    def test_call_00(self):
        tree = FunctionCall(SymbolRef('fn'), [])
        self._check(tree, "fn()")

    def test_call_01(self):
        args = [SymbolRef("arg0")]
        tree = FunctionCall(SymbolRef('fn'), args)
        self._check(tree, "fn(arg0)")

    def test_call_02(self):
        args = [SymbolRef("arg0"), SymbolRef("arg1")]
        tree = FunctionCall(SymbolRef('fn'), args)
        self._check(tree, "fn(arg0, arg1)")



import unittest

from ctree.c.nodes import *


class TestBlocks(unittest.TestCase):
    def _check(self, tree, expected):
        actual = str(tree)
        self.assertEqual(actual, expected)

    def test_simple_00(self):
        body = [Constant(12), Constant('b')]
        tree = Block(body)
        self._check(tree, """\
{
    12;
    'b';
}""")

    def test_nested_00(self):
        a, b = SymbolRef("a"), Constant('b')
        tree = Block([Block([a, b, a]), b, a])
        self._check(tree, """\
{
    {
        a;
        'b';
        a;
    }
    'b';
    a;
}""")



import logging
log = logging.getLogger(__name__)

import threading
import argparse
import Queue as queue # queue in 3.x

from ctree import CONFIG
from ctree.tune import TuningDriver

import opentuner
from opentuner.measurement import MeasurementInterface
from opentuner.resultsdb.models import Result
from opentuner.tuningrunmain import TuningRunMain
from opentuner.search.manipulator import ConfigurationManipulator
from opentuner.measurement.inputmanager import FixedInputManager
from opentuner.api import TuningRunManager


class OpenTunerDriver(TuningDriver):
    """
    Object that interacts with backend tuners. Provides
    a stream of configurations, as well as an interface
    to report on the performance of each.
    """
    def __init__(self, *ot_args, **ot_kwargs):
        """
        Creates communication queues and spawn a thread
        to run the tuning logic.
        """
        super(OpenTunerDriver, self).__init__()
        self._best_config = None
        interface = CtreeMeasurementInterface(self, *ot_args, **ot_kwargs)
        arg_parser = argparse.ArgumentParser(parents=opentuner.argparsers())
        config_args = CONFIG.get("opentuner", "args").split()
        tuner_args = arg_parser.parse_args(config_args)
        self.manager = TuningRunManager(interface, tuner_args)
        self._converged = False

    def _get_configs(self):
        """Get the next configuration to test."""
        timeout = CONFIG.getint("opentuner", "timeout")
        while True:
            self.curr_desired_result = self.manager.get_next_desired_result()
            if self.curr_desired_result is None:
                break
            yield self.curr_desired_result.configuration.data
            print("Best configuration", self.manager.get_best_configuration())

        log.info("exhausted stream of configurations.")
        best_config = self.manager.get_best_configuration()
        assert best_config != None, "No best configuration reported."
        self._converged = True
        while True:
            yield best_config

    def report(self, **kwargs):
        """Report the performance of the most recent configuration."""
        if not self._converged:
            print("Tuning run result:", self.curr_desired_result.configuration.data, kwargs)
            self.manager.report_result(self.curr_desired_result, Result(**kwargs))
            # result = Result(**kwargs)
            # self._results.put_nowait(result)


class CtreeMeasurementInterface(MeasurementInterface):
    """
    Ctree interface to opentuner.
    """
    def __init__(self, driver, *args, **kwargs):
        """
        Create a new measurement interface with knowledge of the
        queues to communicate with the ctree jit module.
        """
        kwargs['input_manager'] = FixedInputManager()
        super(CtreeMeasurementInterface, self).__init__(*args, **kwargs)
        self._ctree_driver = driver

    def manipulator(self):
        """
        Retrieves the tuning space originally provided
        by the specializer.
        """
        return self._manipulator

    def run(self, desired_result, input, limit):
        """
        Build the program specified by 'desired_result' and report
        its performance. With threading, we place the config in the
        output queue, then wait for a Result in the input queue.
        """
        self._ctree_driver._configs.put(desired_result.configuration.data, True)
        return self._ctree_driver._results.get(True)

    def save_final_config(self, configuration):
        """Report best configuration."""
        log.info("best configuration found to be: %s", configuration)
        self._ctree_driver._best_config = configuration.data



__author__ = 'Chick Markley'

import unittest

from ctree.visual.dot_manager import DotManager
from ctree.frontend import get_ast


def square_of(n):
    return n * n

class TestDotManager(unittest.TestCase):
    """
    Difficult to test because of ipython and dot dependencies
    """

    @unittest.skip("difficult to test because of ipython and dot dependencies")
    def test_c_identity(self):
        tree = get_ast(square_of)
        DotManager.run_dot(tree.to_dot())





"""
create specializer projects
basically copies all files and directories from a template.
"""

import sys
import argparse
import ctree



import collections
import shutil
import os

from ctree.tools.generators import builder as Builder

if sys.version_info >= (3, 0, 0): #python 3
    import configparser as ConfigParser
else:
    import ConfigParser


__author__ = 'chick'


def main(*args):
    '''run ctree utility stuff, currently only the project generator'''

    if sys.argv:
        args = sys.argv[1:]

    parser = argparse.ArgumentParser(prog="ctree", description="ctree is a python SEJITS framework")
    parser.add_argument('-sp', '--startproject', help='generate a specializer project')
    parser.add_argument(
        '-wu', '--wattsupmeter', help="start interactive watts up meter shell", action="store_true"
    )
    parser.add_argument('-p', '--port', help="/dev name to use for wattsup meter port")
    parser.add_argument('-v', '--verbose', help='show more debug than you like', action="store_true")
    parser.add_argument('-dc', '--disable_cache', help='disable and delete the persistent cache', action="store_true")
    parser.add_argument('-ec', '--enable_cache', help='enable the persistent cache', action="store_true")
    parser.add_argument('-cc', '--clear_cache', help='clear the persistent cache', action="store_true")
    args = parser.parse_args(args)

    if args.startproject:
        specializer_name = args.startproject

        print ("create project specializer %s" % specializer_name)

        builder = Builder.Builder("create", specializer_name, verbose=args.verbose)
        builder.build(None, None)

    elif args.wattsupmeter:
        from ctree.metrics.watts_up_reader import WattsUpReader

        port = args.port if args.port else WattsUpReader.guess_port()
        meter = WattsUpReader(port_name=port)
        meter.interactive_mode()

    elif args.enable_cache:
        ctree.CONFIG.set("jit", "CACHE", value="True")
        write_success = write_to_config('jit', 'CACHE', True)
        if write_success: print("[SUCCESS] ctree caching enabled.")

    elif args.disable_cache:
        wipe_cache()
        ctree.CONFIG.set("jit", "CACHE", value="False")
        write_success = write_to_config('jit', 'CACHE', False)
        args.clear_cache = True
        if write_success: print("[SUCCESS] ctree caching disabled.")

    elif args.clear_cache:
        wipe_cache()

    else:
        parser.print_usage()

def get_responsible(section, key):
    """
    :param section: Section to search for
    :param key: key to search for
    :return: path of config file responsible for setting
    """
    first = ctree.CFG_PATHS[-1]
    paths = reversed(ctree.CFG_PATHS)
    for path in paths:
        config = ConfigParser.ConfigParser()
        config.read(path)
        if config.has_option(section, key):
            return path
    return first

def write_to_config(section, key, value):
    '''
    This method handles writing to the closest config file to the current
    project, but does not write to the defaults.cfg file in ctree.
    :return: return True if write is successful. False otherwise.
    '''

    if ctree.CFG_PATHS:
        target = get_responsible(section, key)
        config = ConfigParser.ConfigParser()
        config.read(target)
        print(target)
        if not config.has_section(section):
            config.add_section(section)
        config.set(section, key, value)
        with open(target, 'w') as configfile:
            config.write(configfile)
            configfile.close()
        return True
    else:
        print("[FAILURE] No config file detected. Please create a '.ctree.cfg' file in your project directory.")
        return False

def wipe_cache():
    cache_name = os.path.expanduser(ctree.CONFIG.get('jit','COMPILE_PATH'))
    if os.path.isabs(cache_name):
        cache_name = os.path.abspath(cache_name)
    else:
        splitted = cache_name.split(os.sep)
        while splitted:
            first = splitted[0]
            if first == '.':
                splitted.pop(0)
            elif first == '..':
                os.chdir('../')
                splitted.pop(0)
            else:
                cache_name = os.sep.join(splitted)
                break

    wipe_queue = collections.deque([os.path.abspath(p) for p in os.listdir(os.getcwd())])
    while wipe_queue:
        directory = wipe_queue.popleft()
        if not os.path.isdir(directory):
            continue
        if os.path.split(directory)[-1] == cache_name:
            shutil.rmtree(directory)
        else:
            for sub_item in os.listdir(directory):
                wipe_queue.append(os.path.join(directory, sub_item))

if __name__ == '__main__':
    main(sys.argv[1:])



from __future__ import absolute_import

import types
import sys

import ctypes
from ctypes import *

import logging

from ctree import _TYPE_CODEGENERATORS as generators
from ctree import _TYPE_RECOGNIZERS as recognizers

log = logging.getLogger(__name__)


def register_type_codegenerators(codegen_dict):
    """
    Registers routines for generating code for types.

    :param codegen_dict: Maps type classes to functions that
    take an instance of that class and return the corresponding
    string.
    """
    if sys.version_info >= (3, 0):
        existing_keys = generators.keys()
        new_keys = codegen_dict.keys()
    else:
        existing_keys = generators.viewkeys()
        new_keys = codegen_dict.viewkeys()
    intersection = existing_keys & new_keys
    if intersection:
        log.warning("replacing existing type_codegenerator for %s",
                    intersection)

    for genfn in generators.values():
        assert callable(genfn), "Found a non-callable type_codegen: %s" % genfn

    generators.update(codegen_dict)


def register_type_recognizers(typerec_dict):
    """
    Registers routines for getting ctypes objects from Python objects.

    :param typerec_dict: Maps Python classes to functions that
    take an instance of that class and return the corresponding
    ctypes object.
    """
    if sys.version_info >= (3, 0):
        existing_keys = recognizers.keys()
        new_keys = typerec_dict.keys()
    else:
        existing_keys = recognizers.viewkeys()
        new_keys = typerec_dict.viewkeys()
    intersection = existing_keys & new_keys
    if intersection:
        log.warning("replacing existing type_recognizer for %s", intersection)

    for genfn in recognizers.values():
        assert callable(genfn), "Found a non-callable type_codegen: %s" % genfn

    recognizers.update(typerec_dict)


def get_ctype(py_obj):
    """
    Given a python object, this routine tries to return the
    closest ctype type instance corresponding to that object.

    :param py_obj: A python object.
    """
    bases = [type(py_obj)]
    while bases:
        base = bases.pop()
        bases += base.__bases__
        try:
            return recognizers[base](py_obj)
        except KeyError:
            pass
    raise ValueError("No type recognizer defined for %s." % type(py_obj))

def get_c_type_from_numpy_dtype(dtype_specified):
        """
        Get the ctype corresponding to a given numpy.dtype
        :param: dtype_specified - the numpy.dtype
        :return: the ctype corresponding the the dtype_specified. None unable to match
        """
        typemap = {}
        for t in (c_byte, c_short, c_int, c_long, c_longlong):
            typemap["<i%s" % sizeof(t)] = t.__ctype_le__
            typemap[">i%s" % sizeof(t)] = t.__ctype_be__
        for t in (c_ubyte, c_ushort, c_uint, c_ulong, c_ulonglong):
            typemap["<u%s" % sizeof(t)] = t.__ctype_le__
            typemap[">u%s" % sizeof(t)] = t.__ctype_be__
        for t in (c_float, c_double):
            typemap["<f%s" % sizeof(t)] = t.__ctype_le__
            typemap[">f%s" % sizeof(t)] = t.__ctype_be__
        typemap["|b1"] = c_bool
        typemap["|i1"] = c_byte
        typemap["|u1"] = c_ubyte

        if dtype_specified.descr[0][1] in typemap:
            return typemap[dtype_specified.descr[0][1]]
        else:
            return None

def codegen_type(ctype):
    """
    Unparses the given ctype.

    :param ctype: A ctype type instance to be unparsed.
    """
    assert not isinstance(ctype, type), \
        "Expected a ctypes type instance, not %s, (%s):" % (ctype, type(ctype))

    bases = [type(ctype)]
    while bases:
        base = bases.pop()
        bases.extend(base.__bases__)
        try:
            val = generators[base](ctype)
            return val
        except KeyError:
            pass
    raise ValueError("No code generator defined for %s." % type(ctype))


def get_common_ctype(ctypes_list):
    """
    :param ctypes_list: iterable of ctypes
    :return: calculates the proper ctype for coercion of all types, as per

        If either is      long          double the other is promoted to      long          double
        If either is                    double the other is promoted to                    double
        If either is                    float  the other is promoted to                    float
        If either is long long unsigned int    the other is promoted to long long unsigned int
        If either is long long          int    the other is promoted to long long          int
        If either is long      unsigned int    the other is promoted to long      unsigned int
        If either is long               int    the other is promoted to long               int
        if either is           unsigned int    the other is promoted to           unsigned int
        If either is                    int    the other is promoted to                    int
        Both operands are promoted to int
    """

    # lowest ranking takes precedence
    rankings = [ctypes.c_longdouble, ctypes.c_double, ctypes.c_float,
                ctypes.c_long, ctypes.c_uint, ctypes.c_int, ctypes.c_byte,
                ctypes.c_wchar, ctypes.c_char, ctypes.c_bool, ctypes.c_void_p]
    filtered = []
    for c_type in ctypes_list:
        if c_type not in rankings:
            return c_type
        filtered.append(c_type)
    if filtered:
        return min(filtered, key=rankings.index)
    else:
        return ctypes.c_void_p



"""
ctree extensions for OpenCL.
"""

import logging

log = logging.getLogger(__name__)


import pycl

from ctree.types import (
    codegen_type,
    register_type_recognizers,
    register_type_codegenerators,
)

register_type_recognizers({
})

register_type_codegenerators({
    pycl.cl_context:        lambda t: "cl_context",
    pycl.cl_command_queue:  lambda t: "cl_command_queue",
    pycl.cl_kernel:         lambda t: "cl_kernel",
    pycl.cl_mem:            lambda t: "cl_mem",
})


devices_context_queue_map = {}


def get_context_and_queue_from_devices(devices):
    key = tuple(device.vendor_id for device in devices)
    try:
        return devices_context_queue_map[key]
    except KeyError:
        context = pycl.clCreateContext(devices)
        queue = pycl.clCreateCommandQueue(context)
        devices_context_queue_map[key] = (context, queue)
        return devices_context_queue_map[key]



import unittest

from ctree.frontend import get_ast
from fixtures.sample_asts import *


class TestDotGen(unittest.TestCase):
    """
    Run the to_dot function on the fixture ASTs. There isn't
    a good way to check if the output is a legal DOT file,
    but we can at least ensure that the routine runs without
    throwing an exception.
    """

    def test_c_identity(self):
        self.assertNotEqual(identity_ast.to_dot(), "")

    def test_c_gcd(self):
        self.assertNotEqual(gcd_ast.to_dot(), "")

    def test_c_fib(self):
        self.assertNotEqual(fib_ast.to_dot(), "")

    def test_c_l2norm(self):
        self.assertNotEqual(l2norm_ast.to_dot(), "")

    def test_py_identity(self):
        self.assertNotEqual(get_ast(identity).to_dot(), "")

    def test_py_gcd(self):
        self.assertNotEqual(get_ast(gcd).to_dot(), "")

    def test_py_fib(self):
        self.assertNotEqual(get_ast(fib).to_dot(), "")



from ctypes import *

from util import CtreeTest
from ctree.c.nodes import *


class TestFile(CtreeTest):
    def test_simple_00(self):
        foo = SymbolRef("foo", sym_type=c_long())
        bar = FunctionDecl(c_double(), SymbolRef("bar"))
        tree = CFile("myfile", [foo, bar])
        self._check_code(tree, """\
        // <file: myfile.c>
        long foo;
        double bar();
        """)



__author__ = 'nzhang-dev'

import unittest

from ctree.frontend import *;
from ctree.transformations import PyBasicConversions, DeclarationFiller


def fib(n):
    a, b = 0, 1
    k = "hello"
    while n > 0:
        a, b = b, a + b
        n -= 1
    return a


class DeclarationTest(unittest.TestCase):

    def test_fib(self):
        py_ast = get_ast(fib).body[0]
        c_ast = PyBasicConversions().visit(py_ast)
        filled_ast = DeclarationFiller().visit(c_ast)


"""
Just-in-time compilation support.
"""

import abc
import copy
import os
import shutil
import re
import atexit
import ast
import logging
import inspect
import hashlib
import json
from collections import namedtuple
import tempfile


import ctree
from ctree.nodes import Project
from ctree.analyses import VerifyOnlyCtreeNodes
from ctree.util import highlight
from ctree.frontend import get_ast, dump
from ctree.transformations import DeclarationFiller
from ctree.c.nodes import CFile, MultiNode
from ctree.ocl.nodes import OclFile
from ctree.nodes import File

# import llvmlite.binding as llvm
# llvm.initialize()
# llvm.initialize_native_target()

import logging

log = logging.getLogger(__name__)


def getFile(filepath):
    """
    Takes a filepath and returns a specialized File instance (i.e. OclFile, CFile, etc)
    """
    ext_map = {'.'+t._ext:t for t in (
        CFile, OclFile
    )}
    path, filename = os.path.split(filepath)
    name, ext = os.path.splitext(filename)
    filetype = ext_map[ext]
    return filetype(name=name, path=path)


class JitModule(object):
    """
    Manages compilation of multiple ASTs.
    """

    def __init__(self):
        import os

        # write files to $TEMPDIR/ctree/run-XXXX
        ctree_dir = os.path.join(tempfile.gettempdir(), "ctree")
        if not os.path.exists(ctree_dir):
            os.mkdir(ctree_dir)

        self.compilation_dir = tempfile.mkdtemp(prefix="run-", dir=ctree_dir)
        self.ll_module = None
        self.exec_engine = None

    def _link_in(self, submodule):
        self.so_file_name = submodule
        # if self.ll_module is not None:
        #     self.ll_module.link_in(submodule)
        # else:
        #     self.ll_module = submodule

    def get_callable(self, entry_point_name, entry_point_typesig):
        """
        Returns a python callable that dispatches to the requested C function.
        """

        # get llvm represetation of function
        # ll_function = self.ll_module.get_function(entry_point_name)
        import ctypes
        lib = ctypes.cdll.LoadLibrary(self.so_file_name)
        func_ptr = getattr(lib, entry_point_name)
        func_ptr.argtypes = entry_point_typesig._argtypes_
        func_ptr.restype = entry_point_typesig._restype_
        # func = func_ptr

        # run jit compiler
        # from llvm.ee import EngineBuilder
        # self.exec_engine = llvm.create_jit_compiler(self.ll_module)

        # c_func_ptr = self.exec_engine.get_pointer_to_global(ll_function)

        # cast c_func_ptr to python callable using ctypes
        return func_ptr


class ConcreteSpecializedFunction(object):
    """
    A function backed by generated code.
    """
    __metaclass__ = abc.ABCMeta

    def _compile(self, entry_point_name, project_node, entry_point_typesig,
                 **kwargs):
        """
        Returns a python callable.
        """
        assert isinstance(project_node, Project), \
            "Expected a Project but it got a %s." % type(project_node)

        VerifyOnlyCtreeNodes().visit(project_node)

        self._module = project_node.codegen(**kwargs)

        if log.getEffectiveLevel() == 'debug':
            highlighted = highlight(str(self._module.ll_module), 'llvm')
            log.debug("full LLVM program is: <<<\n%s\n>>>" % highlight)

        return self._module.get_callable(entry_point_name, entry_point_typesig)

    @abc.abstractmethod
    def __call__(self, *args, **kwargs):
        pass



class LazySpecializedFunction(object):
    """
    A callable object that will produce executable
    code just-in-time.
    """

    ProgramConfig = namedtuple('ProgramConfig',['args_subconfig', 'tuner_subconfig'])
    _directory_fields = ['__class__.__name__', 'backend_name']

    class NameExtractor(ast.NodeVisitor):
        """
        Extracts the first functiondef name found
        """
        def visit_FunctionDef(self, node):
            return node.name

        def generic_visit(self, node):
            for field, value in ast.iter_fields(node):
                if isinstance(value, list):
                    for item in value:
                        if isinstance(item, ast.AST):
                            res = self.visit(item)
                            if res:
                                return res
                elif isinstance(value, ast.AST):
                    res = self.visit(value)
                    if res:
                        return res

    def __init__(self, py_ast=None, sub_dir=None, backend_name="default"):
        if py_ast is not None and self.apply is not LazySpecializedFunction.apply:
            raise TypeError('Cannot define apply and pass py_ast')
        self.original_tree = py_ast or (get_ast(self.apply) if self.apply is not LazySpecializedFunction.apply else None)
        self.concrete_functions = {}  # config -> callable map
        self._tuner = self.get_tuning_driver()
        self.sub_dir = sub_dir or self.NameExtractor().visit(self.original_tree) or hex(hash(self))[2:]
        self.backend_name = backend_name



    @property
    def original_tree(self):
        return copy.deepcopy(self._original_tree)

    @original_tree.setter
    def original_tree(self, value):
        if not hasattr(self, '_original_tree'):
            self._original_tree = value
        elif ast.dump(self.__original_tree, True, True) != ast.dump(value, True, True):
            raise AttributeError('Cannot redefine the ast')

    @property
    def info_filename(self):
        return 'info.json'

    def get_info(self, path):
        info_filepath = os.path.join(path, self.info_filename)
        if not os.path.exists(info_filepath):
            return {'hash': None, 'files':[]}
        with open(info_filepath) as info_file:
            return json.load(info_file)

    def set_info(self, path, dictionary):
        info_filepath = os.path.join(path, self.info_filename)
        with open(info_filepath,'w') as info_file:
            return json.dump(dictionary, info_file)


    @staticmethod
    def _hash(o):
        if isinstance(o, dict):
            return hash(frozenset(
                LazySpecializedFunction._hash(item) for item in o.items()
            ))
        else:
            return hash(str(o))

    def __hash__(self):
        mro = type(self).mro()
        result = hashlib.sha512(''.encode())
        for klass in mro:
            if issubclass(klass, LazySpecializedFunction):
                try:
                    result.update(inspect.getsource(klass).encode())
                except IOError:  # means source can't be found. Well, can't do anything about that I don't think
                    pass
            else:
                pass
        if self.original_tree is not None:
            tree_str = ast.dump(self.original_tree, annotate_fields=True, include_attributes=True)
            result.update(tree_str.encode())
        return int(result.hexdigest(), 16)


    def config_to_dirname(self, program_config):
        """Returns the subdirectory name under .compiled/funcname"""
        # fixes the directory names and squishes invalid chars
        regex_filter = re.compile(r"""[/\?%*:|"<>()'{} ]""")

        def deep_getattr(obj, s):
            parts = s.split('.')
            for part in parts:
                obj = getattr(obj, part)
            return obj

        path_parts = [
            self.sub_dir,
            str(self._hash(program_config.args_subconfig)),
            str(self._hash(program_config.tuner_subconfig))
            ]

        for attrib in self._directory_fields:
            path_parts.append(str(deep_getattr(self, attrib)))
        filtered_parts = [str(re.sub(regex_filter, '_', part)) for part in path_parts]
        compile_path = str(ctree.CONFIG.get('jit', 'COMPILE_PATH'))

        path = os.path.join(compile_path, *filtered_parts)


        return re.sub('_+','_', path)


    def __call__(self, *args, **kwargs):
        """
            Determines the program_configuration to be run. If it has yet to be
            built, build it. Then, execute it. If the selected program_configuration
            for this function has already been code generated for, this method draws
            from the cache.
        """
        ctree.STATS.log("specialized function call")

        log.info("detected specialized function call with arg types: %s",
                 [type(a) for a in args] + [type(kwargs[key]) for key in kwargs])

        # Don't break old specializers that don't support kwargs
        try:
            args_subconfig = self.args_to_subconfig(args, kwargs)
        except TypeError:
            args_subconfig = self.args_to_subconfig(args)

        tuner_subconfig = next(self._tuner.configs)
        program_config = self.ProgramConfig(args_subconfig, tuner_subconfig)
        dir_name = self.config_to_dirname(program_config)

        if not os.path.exists(dir_name):
            os.makedirs(dir_name)

        log.info("tuner subconfig: %s", tuner_subconfig)
        log.info("arguments subconfig: %s", args_subconfig)

        config_hash = dir_name

        if ctree.CONFIG.getboolean('jit', 'CACHE') and config_hash in self.concrete_functions:              # checks to see if the necessary code is in the run-time cache
            ctree.STATS.log("specialized function cache hit")
            log.info("specialized function cache hit!")
            csf = self.concrete_functions[config_hash]

        else:
            ctree.STATS.log("specialized function cache miss")
            log.info("specialized function cache miss.")
            info = self.get_info(dir_name)

            if hash(self) != info['hash'] and self.original_tree is not None:                      # checks to see if the necessary code is in the persistent cache
                # need to run transform() for code generation
                log.info('Hash miss. Running Transform')
                ctree.STATS.log("Filesystem cache miss")
                transform_result = self.run_transform(program_config)

                # Saving files to cache directory
                for source_file in transform_result:
                    assert isinstance(source_file, File), "Transform must return an iterable of Files"
                    source_file.path = dir_name

                new_info = {'hash': hash(self), 'files':[os.path.join(f.path, f.get_filename()) for f in transform_result]}
                self.set_info(dir_name, new_info)

            else:
                log.info('Hash hit. Skipping transform')
                ctree.STATS.log('Filesystem cache hit')
                files = [getFile(path) for path in info['files']]
                transform_result = files

            csf = self.finalize(transform_result, program_config)
            assert isinstance(csf, ConcreteSpecializedFunction), "Expected a ctree.jit.ConcreteSpecializedFunction, but got a %s." % type(csf)
            self.concrete_functions[config_hash] = csf

        return csf(*args, **kwargs)


    def run_transform(self, program_config):
        transform_result = self.transform(
            self.original_tree,
            program_config
        )
        if not isinstance(transform_result, (tuple, list)):
            transform_result = (transform_result,)

        transform_result = [DeclarationFiller().visit(source_file)
                            if isinstance(source_file, CFile) else source_file
                            for source_file in transform_result]
        return transform_result



    @classmethod
    def from_function(cls, func, folder_name=''):
        class Replacer(ast.NodeTransformer):
            def visit_Module(self, node):
                return MultiNode(body=[self.visit(i) for i in node.body])

            def visit_FunctionDef(self, node):
                if node.name == func.__name__:
                    node.name = 'apply'
                node.body = [self.visit(item) for item in node.body]
                return node

            def visit_Name(self, node):
                if node.id == func.__name__:
                    node.id = 'apply'
                return node

        func_ast = Replacer().visit(get_ast(func))
        return cls(py_ast=func_ast, sub_dir=folder_name or func.__name__)



    def report(self, *args, **kwargs):
        """
        Records the performance of the most recent configuration.
        """
        return self._tuner.report(*args, **kwargs)

    # =====================================================
    # Methods to be overridden by the user

    def transform(self, tree, program_config):
        """
        Convert the AST 'tree' into a C AST, optionally taking advantage of the
        actual runtime arguments.
        """
        raise NotImplementedError()

    def finalize(self, transform_result, program_config):
        """
        This function will be passed the result of transform.  The specializer
        should return an ConcreteSpecializedFunction.
        """
        raise NotImplementedError("Finalize must be implemented")

    def get_tuning_driver(self):
        """
        Define the space of possible implementations.
        """
        from ctree.tune import ConstantTuningDriver

        return ConstantTuningDriver()

    def args_to_subconfig(self, args):
        """
        Extract features from the arguments to define uniqueness of
        this particular invocation. The return value must be a hashable
        object, or a dictionary of hashable objects.
        """
        log.warn("arguments will not influence program_config. " +
                 "Consider overriding args_to_subconfig() in %s.",
                 type(self).__name__)
        return dict()

    @staticmethod
    def apply(*args):
        raise NotImplementedError()



import unittest
import ctypes as ct
import ast
import sys

from ctree.transformations import PyBasicConversions, DeclarationFiller
from ctree.c.nodes import *


class TestAssigns(unittest.TestCase):


    def mini_transform(self, node):
        """
        This method acts as a simulation of a specializer's transform() method. It's the bare minimum required of
        a transform() method by the specializer writer.

        :param node: the node to transform
        :return: the node transformed through PyBasicConversions into a rough C-AST.
        """
        transformed_node = PyBasicConversions().visit(node)

        transformed_node.name = "apply"
        transformed_node.return_type = ct.c_int32()

        for param in transformed_node.params:
            param.type = ct.c_int32()

        return transformed_node

    def mini__call__(self, node):
        """
        This method acts as a simulation of jit.py's __call__() method. The specializer writer does not have to write
        this method.

        :param node: the node to generate code for
        :return: a type-complete C-AST corresponding to the input node
        """
        transformed_node = self.mini_transform(node)
        return DeclarationFiller().visit(transformed_node)

    @unittest.skipIf(sys.version_info >= (3,0), 'Lambdas changed in py3k')
    def test_one_arg_lambda(self):
        """
        This method tests the squaring lambda function, a one argument lambda function.
        """
        square_lambda_node = ast.Lambda(args = ast.arguments([SymbolRef("x")], None, None, None), body = Mul(SymbolRef("x"), SymbolRef("x")))

        # simulating __call__()
        type_inferred_node = self.mini__call__(square_lambda_node)

        self.assertEqual(str(type_inferred_node), "int apply(int x) {\n" + \
                                                "    return x * x;\n}")


    @unittest.skipIf(sys.version_info >= (3,0), 'Lambdas changed in py3k')
    def test_two_arg_lambda(self):
        """
        This method tests the adding lambda function, a two argument lambda function.
        """
        add_lambda_node = ast.Lambda(args = ast.arguments([SymbolRef("x"), SymbolRef("y")], None, None, None), body = Add(SymbolRef("x"), SymbolRef("y")))

        # simulating __call__()
        type_inferred_node = self.mini__call__(add_lambda_node)

        self.assertEqual(str(type_inferred_node), "int apply(int x, int y) {\n" + \
                                                "    return x + y;\n}")



from ctypes import *

from util import CtreeTest
from ctree.c.nodes import *


class TestVarDecls(CtreeTest):
    def test_simple_00(self):
        tree = SymbolRef('foo', c_double())
        self._check_code(tree, "double foo")

    def test_simple_01(self):
        tree = Assign(SymbolRef('foo', c_double()), Constant(1.2))
        self._check_code(tree, "double foo = 1.2")



"""
The ctree package


"""
from __future__ import print_function



# ---------------------------------------------------------------------------
# explicit version check

import sys
# assert sys.version_info[0] >= 3, "ctree requires Python 3.x"
assert sys.version_info[0] >= 2, "ctree requires Python 2.7.x"


# ---------------------------------------------------------------------------
# logging

import logging

LOG = logging.getLogger(__name__)
LOG.info("initializing ctree")

# ---------------------------------------------------------------------------
# configuration file parsing

# pylint disable=import-error
try:
    # python 2
    import ConfigParser as configparser
except ImportError:
    # python 3
    import configparser
# pylint enable=import-error

from os import path, getcwd

CONFIG = configparser.ConfigParser()
DEFAULT_CFG_FILE_PATH = path.join(path.abspath(path.dirname(__file__)), "defaults.cfg")
LOG.info("reading default configuration from: %s", DEFAULT_CFG_FILE_PATH)

CONFIG.readfp(open(DEFAULT_CFG_FILE_PATH), filename="defaults.cfg")

CFG_PATHS = [
    path.expanduser('~/.ctree.cfg'),
    path.join(getcwd(), ".ctree.cfg"),
]
LOG.info("checking for config files at: %s", CFG_PATHS)

LOG.info("found config files: %s", CONFIG.read(CFG_PATHS))

if sys.version_info.major == 2:
    from io import BytesIO as Memfile
else:
    from io import StringIO as Memfile

from ctree.util import highlight

CONFIGFILE = Memfile()
CONFIG.write(CONFIGFILE)
CONFIG_TXT = CONFIGFILE.getvalue()
LOG.info("using configuration:\n%s", highlight(CONFIG_TXT, language='ini'))
CONFIGFILE.close()
if CONFIG.has_option('log','level'):
    logging.basicConfig(level=getattr(logging,CONFIG.get('log','level')))


# ---------------------------------------------------------------------------
# stats

import atexit
import collections


class LogInfo(object):
    """Tracks events, reports counts upon garbage collections."""

    def __init__(self):
        self._counter = collections.Counter()

    def log(self, event_str):
        """record a single named event"""
        self._counter[event_str] += 1

    def report(self):
        """send a counter report of all named events to the log"""
        key_values_string = ""
        for key_value in self._counter.items():
            key_values_string += "  %s: %s\n" % key_value
        LOG.info("execution statistics: (((\n%s)))", key_values_string)


STATS = LogInfo()
atexit.register(STATS.report)

#----------------------------------------------------------------------------
#Temporary directory stuff
import tempfile
import shutil

if CONFIG.getboolean('jit', 'CACHE'):
    STATS.log("recognized that caching is enabled")
else:
    STATS.log("recognized that caching is disabled")

if not CONFIG.getboolean('jit', 'CACHE'):
    compile_path_old = CONFIG.get('jit', 'COMPILE_PATH')
    temporary_path = tempfile.mkdtemp()
    CONFIG.set('jit', 'COMPILE_PATH', temporary_path)

    def reset():
        CONFIG.set('jit', 'COMPILE_PATH', compile_path_old)
        shutil.rmtree(temporary_path)

    atexit.register(reset)

# Registries for type-based logic in extension packages.
_TYPE_CODEGENERATORS = {}
_TYPE_RECOGNIZERS = {}

import ast
import inspect
import ctree.frontend
from ctree.visual.dot_manager import DotManager


def get_ast(func):
    """convenience method for displaying a callable objects ast"""
    return ctree.frontend.get_ast(func)


def ipython_show_ast(tree):
    """
    convenience method to display an AST in ipython
    converts tree in place to a dot format
    then renders that into a png file
    """
    import ctree.dotgen
    return DotManager.dot_ast_to_image(tree)


def browser_show_ast(tree, file_name=None):
    """
    convenience method to display an AST in ipython
    converts tree in place to a dot format
    then renders that into a png file
    """
    import ctree.dotgen
    return DotManager.dot_ast_to_browser(tree, file_name)



import unittest

from ctree.analyses import *
from ctree.frontend import get_ast
from fixtures.sample_asts import *


class TestVerifyOnlyCtreeNodes(unittest.TestCase):
    def _check(self, tree):
        VerifyOnlyCtreeNodes().visit(tree)

    def _check_raises(self, tree):
        with self.assertRaises(AstValidationError):
            self._check(tree)

    def test_identity_py(self):
        self._check_raises(get_ast(identity))

    def test_fib_py(self):
        self._check_raises(get_ast(fib))

    def test_gcd_py(self):
        self._check_raises(get_ast(gcd))

    def test_identity_c(self):
        self._check(identity_ast)

    def test_fib_c(self):
        self._check(fib_ast)

    def test_gcd_c(self):
        self._check(gcd_ast)



import unittest

from ctree.c.macros import *


class TestCMacrosCodegen(unittest.TestCase):
    def test_null(self):
        node = NULL()
        self.assertEqual(str(node), "NULL")

    def test_printf(self):
        node = printf("%s %s", SymbolRef("x"), SymbolRef("y"))
        self.assertEqual(str(node), "printf(\"%s %s\", x, y)")



"""
validation of ast trees
"""
from ctree.visitors import NodeVisitor
from ctree.nodes import CtreeNode, ast
from ctree.c.nodes import SymbolRef
from ctree.util import flatten


class DeclFinder(NodeVisitor):
    """
    Returns the first use of a particular symbol.
    """

    def __init__(self):
        self.decl = None

    def find(self, node):
        """look for a declaration of a symbol ref"""
        assert isinstance(node, SymbolRef), \
            "DeclFinder only works on SymbolRefs for now."

        self.decl = None

        assert self.decl is not None, \
            "Couldn't find declaration for symbol %s." % node
        return self.decl


class AstValidationError(Exception):
    """
    Exception for non C nodes in an AST
    """
    pass


class VerifyOnlyCtreeNodes(NodeVisitor):
    """
    Checks that every node in the tree is an instance of
    ctree.nodes.common.CtreeNode. Raises an exception if a bad node
    is found.
    """

    def visit(self, node):
        if not isinstance(node, CtreeNode):
            raise AstValidationError("Expected a pure C ast, but found a non-CtreeNode: %s." % node)
        self.generic_visit(node)



import unittest

from ctree.c.nodes import *


class TestUnaryOps(unittest.TestCase):
    def setUp(self):
        self.foo = SymbolRef("foo")

    def _check(self, op_cls, expected_string):
        node = op_cls(self.foo)
        self.assertEqual(str(node), expected_string)

    def test_plus(self):
        self._check(Add, "+ foo")

    def test_minus(self):
        self._check(Sub, "- foo")

    def test_bitnot(self):
        self._check(BitNot, "~ foo")

    def test_not(self):
        self._check(Not, "! foo")

    def test_ref(self):
        self._check(Ref, "& foo")

    def test_deref(self):
        self._check(Deref, "* foo")

    def test_preinc(self):
        self._check(PreInc, "++ foo")

    def test_predec(self):
        self._check(PreDec, "-- foo")

    def test_postinc(self):
        self._check(PostInc, "foo ++")

    def test_postdec(self):
        self._check(PostDec, "foo --")

    def test_sizeof(self):
        self._check(SizeOf, "sizeof foo")



import unittest

from fixtures.sample_asts import *
from ctree.transformations import *
from ctree.c.nodes import *
from ctree.frontend import get_ast


class TestCtxScrubber(unittest.TestCase):
    def _check(self, tree):
        for node in ast.walk(tree):
            if isinstance(node, ast.Name):
                self.assertIsNone(node.ctx)

    def test_identity(self):
        tree = PyCtxScrubber().visit(get_ast(identity))
        self._check(tree)

    def test_fib(self):
        tree = PyCtxScrubber().visit(get_ast(fib))
        self._check(tree)

    def test_gcd(self):
        tree = PyCtxScrubber().visit(get_ast(gcd))
        self._check(tree)


class TestStripDocstrings(unittest.TestCase):
    def foo(self):
        """my docstring"""
        pass

    class Bar():
        """my other docstring"""
        pass

    def _check(self, tree):
        for node in ast.walk(tree):
            self.assertNotIsInstance(node, ast.Str)

    def test_func_docstring(self):
        tree = PyBasicConversions().visit(get_ast(self.foo))
        self._check(tree)

    def test_class_docstring(self):
        tree = PyBasicConversions().visit(get_ast(self.Bar))
        self._check(tree)

    def test_subtree_docstrings(self):
        tree = PyBasicConversions().visit(ast.Module([
            get_ast(self.foo),
            get_ast(self.Bar),
        ]))
        self._check(tree)

class TestBasicConversions(unittest.TestCase):
    def _check(self, py_ast, expected_c_ast, names_dict ={}, constants_dict={}):
        actual_c_ast = PyBasicConversions(names_dict, constants_dict).visit(py_ast)
        self.assertEqual(str(actual_c_ast).strip('\n;'), str(expected_c_ast).strip('\n;'))


    def test_num_float(self):
        py_ast = ast.Num(123.4)
        c_ast = Constant(123.4)
        self._check(py_ast, c_ast)

    def test_num_int(self):
        py_ast = ast.Num(123)
        c_ast = Constant(123)
        self._check(py_ast, c_ast)

    def test_str(self):
        py_ast = ast.Str("foo bar")
        c_ast = String("foo bar")
        self._check(py_ast, c_ast)

    def test_name(self):
        py_ast = ast.Name("baz", ast.Load())
        c_ast = SymbolRef("baz")
        self._check(py_ast, c_ast)

    def test_binop(self):
        for py_op, c_op in (
                (ast.Add, Add),
                (ast.Sub, Sub),
                (ast.BitXor, BitXor),
                (ast.BitAnd, BitAnd),
                (ast.BitOr, BitOr)
        ):
            py_ast = ast.BinOp(ast.Num(1), py_op(), ast.Num(2))
            c_ast = c_op(Constant(1), Constant(2))
            self._check(py_ast, c_ast)

    def test_return(self):
        py_ast = ast.Return()
        c_ast = Return()
        self._check(py_ast, c_ast)

    def test_if_else(self):
        py_ast = ast.If(ast.Num(1), [ast.Num(2)], [ast.Num(3)])
        c_ast = If(Constant(1), [Constant(2)], [Constant(3)])
        self._check(py_ast, c_ast)

    def test_if(self):
        py_ast = ast.If(ast.Num(1), [ast.Num(2)], [])
        c_ast = If(Constant(1), [Constant(2)])
        self._check(py_ast, c_ast)

    def test_ifexp(self):
        py_ast = ast.IfExp(ast.Num(1), ast.Num(2), ast.Num(3))
        c_ast = TernaryOp(Constant(1), Constant(2), Constant(3))
        self._check(py_ast, c_ast)

    def test_compare(self):
        py_ast = ast.Compare(ast.Num(1), [ast.Lt()], [ast.Num(2)])
        c_ast = Lt(Constant(1), Constant(2))
        self._check(py_ast, c_ast)

    @unittest.skipIf(sys.version_info.major == 2, "arg not defined for python2")
    def test_arg(self):
        py_ast = ast.arg("foo", None)
        c_ast = SymbolRef("foo")
        self._check(py_ast, c_ast)

    def test_for_1_arg(self):
        stop = ast.Num(10)
        py_ast = ast.For(ast.Name("i", ast.Load()),
            ast.Call(ast.Name("range", ast.Load()), [stop], [], None, None),
            [ast.Name("foo", ast.Load())],
            [],
        )
        i = SymbolRef("i", c_long())
        c_ast = For(
            Assign(i, Constant(0)),
            Lt(i.copy(), Constant(10)),
            AddAssign(i.copy(), Constant(1)),
            [SymbolRef("foo")],
        )
        self._check(py_ast, c_ast)

    def test_for_2_args(self):
        start = ast.Num(2)
        stop = ast.Num(10)
        py_ast = ast.For(ast.Name("i", ast.Load()),
            ast.Call(ast.Name("range", ast.Load()), [start, stop], [], None, None),
            [ast.Name("foo", ast.Load())],
            [],
        )
        i = SymbolRef("i", c_long())
        c_ast = For(
            Assign(i, Constant(2)),
            Lt(i.copy(), Constant(10)),
            AddAssign(i.copy(), Constant(1)),
            [SymbolRef("foo")],
        )
        self._check(py_ast, c_ast)

    def test_for_3_args(self):
        start = ast.Num(2)
        stop = ast.Num(10)
        step = ast.Num(3)
        py_ast = ast.For(ast.Name("i", ast.Load()),
            ast.Call(ast.Name("range", ast.Load()), [start, stop, step], [], None, None),
            [ast.Name("foo", ast.Load())],
            [],
        )
        i = SymbolRef("i", c_long())
        c_ast = For(
            Assign(i, Constant(2)),
            Lt(i.copy(), Constant(10)),
            AddAssign(i.copy(), Constant(3)),
            [SymbolRef("foo")],
        )
        self._check(py_ast, c_ast)

    def test_for_0_args(self):
        py_ast = ast.For(ast.Name("i", ast.Load()),
            ast.Call(ast.Name("range", ast.Load()), [], [], None, None),
            [ast.Name("foo", ast.Load())],
            [],
        )
        with self.assertRaises(Exception):
          self._check(py_ast, None)

    def test_for_4_args(self):
        py_ast = ast.For(ast.Name("i", ast.Load()),
            ast.Call(ast.Name("range", ast.Load()),
                [Constant(1), Constant(2), Constant(3), Constant(4)], [], None, None),
            [ast.Name("foo", ast.Load())],
            [],
        )
        with self.assertRaises(Exception):
          self._check(py_ast, None)

    def test_for_expr_args(self):
        start = ast.BinOp(ast.Num(2), ast.Add(), ast.Num(3))
        stop = ast.BinOp(ast.Num(4), ast.Mult(), ast.Num(10))
        step = ast.BinOp(ast.Num(3), ast.Mod(), ast.Num(5))
        py_ast = ast.For(ast.Name("i", ast.Load()),
            ast.Call(ast.Name("range", ast.Load()), [start, stop, step], [], None, None),
            [ast.Name("foo", ast.Load())],
            [],
        )
        i = SymbolRef("i", c_long())
        c_ast = For(
            Assign(i, Add(Constant(2), Constant(3))),
            Lt(i.copy(), Mul(Constant(4), Constant(10))),
            AddAssign(i.copy(), Mod(Constant(3), Constant(5))),
            [SymbolRef("foo")],
        )
        self._check(py_ast, c_ast)

    def test_AddAssign(self):
        py_ast = ast.AugAssign(ast.Name('i', ast.Load()), ast.Add(), ast.Num(3))
        c_ast = AddAssign(SymbolRef('i'), Constant(3))
        self._check(py_ast, c_ast)

    def test_SubAssign(self):
        py_ast = ast.AugAssign(ast.Name('i', ast.Load()),
                               ast.Sub(), ast.Num(3))
        c_ast = SubAssign(SymbolRef('i'), Constant(3))
        self._check(py_ast, c_ast)

    def test_MulAssign(self):
        py_ast = ast.AugAssign(ast.Name('i', ast.Load()),
                               ast.Mult(), ast.Num(3))
        c_ast = MulAssign(SymbolRef('i'), Constant(3))
        self._check(py_ast, c_ast)

    def test_DivAssign(self):
        py_ast = ast.AugAssign(ast.Name('i', ast.Load()),
                               ast.Div(), ast.Num(3))
        c_ast = DivAssign(SymbolRef('i'), Constant(3))
        self._check(py_ast, c_ast)

    def test_AugAssign(self):

        for py_op, c_op in (
                (
                        (ast.Div, DivAssign),
                        (ast.Add, AddAssign),
                        (ast.Mult, MulAssign),
                        (ast.BitOr, BitOrAssign),
                        (ast.BitAnd, BitAndAssign),
                        (ast.BitXor, BitXorAssign),
                        (ast.LShift, BitShLAssign),
                        (ast.RShift, BitShRAssign)
                )
        ):
            py_ast = ast.AugAssign(ast.Name('i', ast.Load()),
                                   py_op(), ast.Num(3))
            c_ast = c_op(SymbolRef('i'), Constant(3))
            self._check(py_ast, c_ast)

    def test_Assign(self):
        py_ast = ast.Assign([ast.Name('i', ast.Load())],
                            ast.Num(3))
        c_ast = Assign(SymbolRef('i'), Constant(3))
        self._check(py_ast, c_ast)

    def test_namesDict(self):
        py_ast = ast.Name('i',ast.Load())
        c_ast = SymbolRef('d')
        self._check(py_ast,c_ast,names_dict={'i':'d'})

    def test_constantsDict(self):
        py_ast = ast.Name('i',ast.Load())
        c_ast = Constant(234)
        self._check(py_ast,c_ast,constants_dict={'i':234})

    def test_Subscript(self):
        py_ast = ast.Subscript(value=ast.Name('i',ast.Load()),
                               slice=ast.Index(value=ast.Num(n=1), ctx=ast.Load()))
        c_ast = ArrayRef(SymbolRef('i'),Constant(1))
        self._check(py_ast,c_ast)

    def test_Range_ValueError(self):
        py_ast = ast.For(target=ast.Name(id='i', ctx=ast.Store()), iter=ast.Call(func=ast.Name(id='range', ctx=ast.Load()), args=[
            ast.Num(n=1),
            ast.Num(n=0),
            ast.Num(n=0),
            ], keywords=[], starargs=None, kwargs=None), body=[
            Pass(),
            ], orelse=[])
        with self.assertRaises(ValueError):
            PyBasicConversions().visit(py_ast)

    def test_Range_NoOp(self):
        py_ast = ast.For(target=ast.Name(id='i', ctx=ast.Store()), iter=ast.Call(func=ast.Name(id='range', ctx=ast.Load()), args=[
            ast.Num(n=1),
            ast.Num(n=1),
            ast.Num(n=3),
            ], keywords=[], starargs=None, kwargs=None), body=[
            Pass(),
            ], orelse=[])
        self.assertEqual(PyBasicConversions().visit(py_ast), None)



"""
Macros for simplifying construction of C programs.
"""

from ctree.c.nodes import SymbolRef, CtreeNode, FunctionCall, String


def NULL():
    """
    The NULL symbol.
    """
    return SymbolRef("NULL")


def printf(fmt, *args):
    """
    Makes a printf call. Args must be CtreeNodes.
    """
    for arg in args:
        assert isinstance(arg, CtreeNode)
    return FunctionCall(SymbolRef("printf"),
                        [String(fmt)] + list(args))



import unittest

from ctree.nodes import *
from ctree.jit import LazySpecializedFunction
from ctree.jit import ConcreteSpecializedFunction
from ctree.frontend import dump
from fixtures.sample_asts import *
import ctypes


class TestTranslator(LazySpecializedFunction):
    def args_to_subconfig(self, args):
        return {'arg_typesig': tuple(type(get_ctype(a)) for a in args)}

    def transform(self, tree, program_config):
        arg_types = program_config[0]['arg_typesig']

        tree.return_type = arg_types[0]()
        for param, ty in zip(tree.params, arg_types):
            param.type = ty()
        return [CFile(tree.name, [tree])]

    def finalize(self, transform_result, program_config):
        proj = Project(transform_result)
        cfile = transform_result[0]
        arg_types = program_config[0]['arg_typesig']

        func_type = ctypes.CFUNCTYPE(arg_types[0], *arg_types)
        return BasicFunction(cfile.name, proj, func_type)


class BasicFunction(ConcreteSpecializedFunction):
    def __init__(self, entry, tree, typesig):
        self._c_function = self._compile(entry, tree, typesig)

    def __call__(self, *args, **kwargs):
        return self._c_function(*args, **kwargs)


class BadArgs(LazySpecializedFunction):
    def args_to_subconfig(self, args):
        return {'args': args}


class DefaultArgs(LazySpecializedFunction):
    def transform(self, tree, program_config):
        return CFile("generated", [tree])

    def finalize(self, transform_result, program_config):
        proj = Project(transform_result)
        ctype = self.original_tree.get_type().as_ctype()
        return BasicFunction(self.original_tree.name, proj, ctype)


class NoTransform(LazySpecializedFunction):
    def args_to_subconfig(self, args):
        return {'arg_typesig': tuple(type(get_ctype(arg)) for arg in args)}


#@unittest.skip('Removed Support for AST injection')
class TestSpecializers(unittest.TestCase):
    def test_identity_int(self):
        c_identity = TestTranslator(identity_ast, sub_dir='test_identity_int')
        self.assertEqual(c_identity(1), identity(1))

    def test_identity_float(self):
        c_identity = TestTranslator(identity_ast, sub_dir='test_identity_float')
        self.assertEqual(c_identity(1.2), identity(1.2))

    
    def test_fib_int(self):
        c_fib = TestTranslator(fib_ast, sub_dir='test_fib_int')
        self.assertEqual(c_fib(1), fib(1))

    
    def test_fib_float(self):
        c_fib = TestTranslator(fib_ast, sub_dir='test_fib_float')
        self.assertEqual(c_fib(1.2), fib(1.2))

    
    def test_fib_intfloat(self):
        c_fib = TestTranslator(fib_ast, 'test_fib_intfloat')
        self.assertEqual(c_fib(1), fib(1))
        self.assertEqual(c_fib(1.2), fib(1.2))
    
    def test_gcd_int(self):
        c_gcd = TestTranslator(gcd_ast, 'test_gcd_int')
        self.assertEqual(c_gcd(1, 2), gcd(1, 2))
    
    def test_default_args_to_subconfig(self):
        c_identity = DefaultArgs(identity_ast, 'test_default_args_to_subconfig')
        self.assertEqual(c_identity.args_to_subconfig([1, 2, 3]), {})
    
    def test_no_transform(self):
        c_identity = NoTransform(identity_ast, 'test_no_transform')
        with self.assertRaises(NotImplementedError):
            self.assertEqual(c_identity(1.2), identity(1.2))
    
    def test_repeated(self):
        c_fib = TestTranslator(fib_ast, 'test_repeated')
        for i in range(20):
            self.assertEqual(c_fib(1), fib(1))

if __name__ == '__main__':
    unittest.main()


"""
A set of basic transformers for python asts
"""
import os
import sys
import ast
from ctypes import c_long, c_int, c_byte, c_short, c_char_p, c_void_p
import ctypes
from collections import deque

import ctree
from ctree.c.nodes import Constant, String, SymbolRef, BinaryOp, TernaryOp, \
    Return, While, MultiNode, UnaryOp
from ctree.c.nodes import If, CFile, FunctionCall, FunctionDecl, For, Assign, \
    ArrayRef
from ctree.nodes import CtreeNode
from ctree.c.nodes import Lt, Gt, AddAssign
from ctree.c.nodes import Break, Continue, Pass, Array, Literal, And
from ctree.c.nodes import Op
from ctree.visitors import NodeTransformer

from ctree.types import get_common_ctype


#  conditional imports

if sys.version_info < (3, 0):
    from itertools import izip_longest
else:
    from itertools import zip_longest as izip_longest


def get_type(node):
    if hasattr(node, 'get_type'):
        return type(node.get_type())
    elif hasattr(node, 'type'):
        return type(node.type)
    return c_void_p


class PyCtxScrubber(NodeTransformer):
    """
    Removes pesky ctx attributes from Python ast.Name nodes,
    yielding much cleaner python asts.
    """

    def visit_Name(self, node):
        node.ctx = None
        return node


class PyBasicConversions(NodeTransformer):
    """
    Convert constructs with obvious C analogues.
    """
    def __init__(self, names_dict={}, constants_dict={}):
        self.names_dict = names_dict
        self.constants_dict = constants_dict

    PY_OP_TO_CTREE_OP = {
        ast.Add: Op.Add,
        ast.Mod: Op.Mod,
        ast.Mult: Op.Mul,
        ast.Sub: Op.Sub,
        ast.Div: Op.Div,
        ast.Lt: Op.Lt,
        ast.Gt: Op.Gt,
        ast.LtE: Op.LtE,
        ast.GtE: Op.GtE,
        ast.BitAnd: Op.BitAnd,
        ast.BitOr: Op.BitOr,
        ast.Eq: Op.Eq,
        ast.NotEq: Op.NotEq,
        ast.Not: Op.Not,
        ast.And: Op.And,
        ast.Or: Op.Or,
        ast.BitXor: Op.BitXor,
        ast.LShift: Op.BitShL,
        ast.RShift: Op.BitShR,
        ast.Is: Op.Eq,
        ast.IsNot: Op.NotEq,
        ast.USub: Op.SubUnary,
        ast.UAdd: Op.AddUnary,
        ast.FloorDiv: Op.Div
        # TODO list the rest
    }

    PY_UOP_TO_CTREE_UOP = {
        'UAdd': Op.Add,
        'USub': Op.Sub,
        'Not': Op.Not,
        'Invert': Op.BitNot
    }

    def visit_Num(self, node):
        return Constant(node.n)

    def visit_Str(self, node):
        return String(node.s)

    def visit_Name(self, node):
        if node.id in self.constants_dict:
            return Constant(self.constants_dict[node.id])
        if node.id in self.names_dict:
            return SymbolRef(self.names_dict[node.id])
        return SymbolRef(node.id)

    def visit_BinOp(self, node):
        lhs = self.visit(node.left)
        rhs = self.visit(node.right)
        op = self.PY_OP_TO_CTREE_OP.get(type(node.op), type(node.op))()
        return BinaryOp(lhs, op, rhs)

    def visit_Return(self, node):
        if hasattr(node, 'value'):
            return Return(self.visit(node.value))
        else:
            return Return()

    def visit_For(self, node):
        """restricted, for now, to range as iterator with long-type args"""
        if isinstance(node, ast.For) and \
           isinstance(node.iter, ast.Call) and \
           isinstance(node.iter.func, ast.Name) and \
           node.iter.func.id in ('range', 'xrange'):
            Range = node.iter
            nArgs = len(Range.args)
            if nArgs == 1:
                stop = self.visit(Range.args[0])
                start, step = Constant(0), Constant(1)
            elif nArgs == 2:
                start, stop = map(self.visit, Range.args)
                step = Constant(1)
            elif nArgs == 3:
                start, stop, step = map(self.visit, Range.args)
            else:
                raise Exception(
                    "Cannot convert a for...range with %d args." % nArgs)

            #  check no-op conditions.
            if all(isinstance(item, Constant) for item in (start, stop, step)):
                if step.value == 0:
                    raise ValueError("range() step argument must not be zero")
                elif start.value == stop.value or \
                        (start.value < stop.value and step.value < 0) or \
                        (start.value > stop.value and step.value > 0):
                    return None

            if not all(isinstance(item, CtreeNode)
                       for item in (start, stop, step)):
                return node

            # TODO allow any expressions castable to Long type
            target_types = [c_long]
            for el in (stop, start, step):
                #  typed item to try and guess type off of. Imperfect right now.
                if hasattr(el, 'get_type'):
                    # TODO take the proper class instead of the last; if start,
                    # end are doubles, but step is long, target is double
                    t = el.get_type()
                    assert any(isinstance(t, klass) for klass in [
                        c_byte, c_int, c_long, c_short
                    ]), "Can only convert ranges with integer/long \
                         start/stop/step values"
                    target_types.append(type(t))
            target_type = get_common_ctype(target_types)()

            target = SymbolRef(node.target.id, target_type)
            op = Lt
            if hasattr(start, 'value') and hasattr(stop, 'value') and \
                    start.value > stop.value:
                op = Gt
            for_loop = For(
                Assign(target, start),
                op(target.copy(), stop),
                AddAssign(target.copy(), step),
                [self.visit(stmt) for stmt in node.body],
            )
            return for_loop
        node.body = list(map(self.visit, node.body))
        return node

    def visit_If(self, node):
        if isinstance(node, ast.If):
            cond = self.visit(node.test)
            then = [self.visit(t) for t in node.body]
            elze = [self.visit(t) for t in node.orelse] or None
            return If(cond, then, elze)
        else:
            return self.generic_visit(node)

    def visit_IfExp(self, node):
        cond = self.visit(node.test)
        then = self.visit(node.body)
        elze = self.visit(node.orelse)
        return TernaryOp(cond, then, elze)

    def visit_BoolOp(self, node):
        first = self.visit(node.values[0])
        second = self.visit(node.values[1])
        op = self.PY_OP_TO_CTREE_OP.get(type(node.op),
                                        type(node.op))()
        curr = BinaryOp(first, op, second)
        for value in node.values[2:]:
            curr = BinaryOp(curr, op, self.visit(value))
        return curr

    def visit_Compare(self, node):
        lhs = self.visit(node.left)

        op = self.PY_OP_TO_CTREE_OP.get(type(node.ops[0]),
                                        type(node.ops[0]))()
        rhs = self.visit(node.comparators[0])
        curr = BinaryOp(lhs, op, rhs)
        for i in range(1, len(node.ops)):
            op = self.PY_OP_TO_CTREE_OP.get(type(node.ops[i]),
                                            type(node.ops[i]))()
            rhs = self.visit(node.comparators[i])
            lhs = self.visit(node.comparators[i-1])
            curr = And(curr, BinaryOp(lhs, op, rhs))
        return curr

    def visit_Module(self, node):
        body = [self.visit(s) for s in node.body]
        return CFile("module", body)

    def visit_Call(self, node):
        args = [self.visit(a) for a in node.args]
        fn = self.visit(node.func)
        return FunctionCall(fn, args)

    def visit_FunctionDef(self, node):
        if ast.get_docstring(node):
            node.body.pop(0)
        params = [self.visit(p) for p in node.args.args]
        defn = [self.visit(s) for s in node.body]
        return FunctionDecl(None, node.name, params, defn)

    def visit_arg(self, node):
        return SymbolRef(node.arg, node.annotation)

    def visit_AugAssign(self, node):
        op = type(node.op)
        target = self.visit(node.target)
        value = self.visit(node.value)
        # if op is ast.Add:
        #     return AddAssign(target, value)
        # elif op is ast.Sub:
        #     return SubAssign(target, value)
        # elif op is ast.Mult:
        #     return MulAssign(target, value)
        # elif op is ast.Div:
        #     return DivAssign(target, value)
        # elif op is ast.BitXor:
        #     return BitXorAssign(target, value)
        # # TODO: Error?
        lookup = {
            ast.Add: 'AddAssign', ast.Sub: 'SubAssign', ast.Mult: 'MulAssign',
            ast.Div: 'DivAssign', ast.BitAnd: 'BitAndAssign', ast.BitOr:
            'BitOrAssign', ast.BitXor: 'BitXorAssign', ast.LShift:
            'BitShLAssign', ast.RShift: 'BitShRAssign'
        }
        if op in lookup:
            return getattr(ctree.c.nodes, lookup[op])(target, value)
        return node

    def targets_to_list(self, targets):
        """parses target into nested lists"""
        res = []
        for elt in targets:
            if not isinstance(elt, (ast.List, ast.Tuple)):
                res.append(elt)
            elif isinstance(elt, (ast.Tuple, ast.List)):
                res.append(self.targets_to_list(elt.elts))
        return res

    def value_to_list(self, value):
        """parses value into nested lists for multiple assign"""
        res = []
        if not isinstance(value, (ast.List, ast.Tuple)):
            return value
        for elt in value.elts:
            if not isinstance(value, (ast.List, ast.Tuple)):
                res.append(elt)
            else:
                res.append(self.value_to_list(elt))
        return ast.List(elts=res)

    def pair_lists(self, targets, values):
        res = []
        queue = deque((target, values) for target in targets)
        sentinel = object()
        while queue:
            target, value = queue.popleft()
            if isinstance(target, list):
                #  target hasn't been completely unrolled yet
                for sub_target, sub_value in izip_longest(
                        target, value.elts, fillvalue=sentinel):
                    if sub_target is sentinel or \
                            sub_value is sentinel:
                        raise ValueError(
                            'Incorrect number of values to unpack')
                    queue.append((sub_target, sub_value))
            else:
                res.append((target, value))
        return res

    def parse_pairs(self, node):
        targets = self.targets_to_list(node.targets)
        values = self.value_to_list(node.value)
        return self.pair_lists(targets, values)

    def visit_Assign(self, node):

        target_value_list = [(self.visit(target), self.visit(value))
                             for target, value in self.parse_pairs(node)]

        # making a multinode no matter what. It's cleaner than branching a lot
        operation_body = []
        swap_body = []
        for target, value in target_value_list:
            if not isinstance(target, SymbolRef):
                operation_body.append(Assign(target, value))
            elif isinstance(value, Literal) and \
                    not isinstance(value, SymbolRef):
                operation_body.append(Assign(target, value))
            else:
                new_target = target.copy()
                new_target.name = "____temp__" + new_target.name
                operation_body.append(Assign(new_target, value))
                swap_body.append(Assign(target, new_target.copy()))
        return MultiNode(body=operation_body + swap_body)

    def visit_Subscript(self, node):
        if isinstance(node.slice, ast.Index):
            value = self.visit(node.value)
            index = self.visit(node.slice.value)
            return ArrayRef(value, index)
        else:
            return node

    def visit_While(self, node):
        cond = self.visit(node.test)
        body = [self.visit(i) for i in node.body]
        return While(cond, body)

    def visit_Lambda(self, node):

        if isinstance(node, ast.Lambda):
            def_node = ast.FunctionDef(name="default", args=node.args,
                                       body=node.body, decorator_list=None)

            params = [self.visit(p) for p in def_node.args.args]
            defn = [Return(self.visit(def_node.body))]
            decl_node = FunctionDecl(None, def_node.name, params, defn)
            Lifter().visit_FunctionDecl(decl_node)

            return decl_node
        else:
            return node

    def visit_Break(self, node):
        return Break()

    def visit_Continue(self, node):
        return Continue()

    def visit_Pass(self, node):
        return Pass()

    def visit_List(self, node):
        elts = [self.visit(elt) for elt in node.elts]
        types = [get_type(elt) for elt in elts]
        array_type = get_common_ctype(types)
        return Array(type=ctypes.POINTER(array_type)(), body=elts)

    def visit_UnaryOp(self, node):
        argument = self.visit(node.operand)
        op = self.PY_OP_TO_CTREE_OP.get(type(node.op), type(node.op))()
        return UnaryOp(op, argument)


class ResolveGeneratedPathRefs(NodeTransformer):
    """
    Converts any instances of ctree.nodes.GeneratedPathRef into strings
    containing the absolute path of the target file.
    """

    def __init__(self, compilation_dir):
        self.compilation_dir = compilation_dir
        self.count = 0

    def visit_GeneratedPathRef(self, node):
        self.count += 1
        return String(os.path.join(self.compilation_dir,
                                   node.target.get_filename()))


class Lifter(NodeTransformer):
    """
    To aid in adding new includes or parameters during tree
    traversals, users can store them with arbitrary child nodes and call this
    transformation to move them to the correct position.
    """
    def __init__(self, lift_params=True, lift_includes=True):
        self.lift_params = lift_params
        self.lift_includes = lift_includes

    def visit_FunctionDecl(self, node):
        if self.lift_params:
            for child in ast.walk(node):
                for param in getattr(child, '_lift_params', []):
                    if param not in node.params:
                        node.params.append(param)
                    #  del child._lift_params
        return self.generic_visit(node)

    def visit_CFile(self, node):
        if self.lift_includes:
            new_includes = []
            for child in ast.walk(node):
                for include in getattr(child, '_lift_includes', []):
                    if include not in new_includes:
                        new_includes.append(include)
            node.body = list(new_includes) + node.body
        return self.generic_visit(node)


class DeclarationFiller(NodeTransformer):
    def __init__(self):
        self.__environments = [{}]

    def _lookup(self, key):
        """
        :param key:
        :return: Looks up the last value corresponding to key in
            self.__environments
        """
        if isinstance(key, SymbolRef):
            key = key.name
        value = sentinel = object()
        for environment in self.__environments:
            if key in environment:
                value = environment[key]
        if value is sentinel:
            raise KeyError('Did not find {} in environments'.format(repr(key)))
        return value

    def _has_key(self, key):
        try:
            self._lookup(key)
            return True
        except KeyError:
            return False

    def __add_entry(self, key, value):
        if isinstance(key, SymbolRef):
            key = key.name
        self.__environments[-1][key] = value

    def __add_environment(self):
        self.__environments.append({})

    def __pop_environment(self):
        return self.__environments.pop()

    def visit_FunctionDecl(self, node):
        # add current FunctionDecl's return type onto environments
        self.__add_entry(node.name, node.return_type)

        # new environment every time we enter a function
        self.__add_environment()

        for param in node.params:
            # binding types of parameters
            self.__add_entry(param.name, param.type)

        node.defn = [self.visit(i) for i in node.defn]
        self.__pop_environment()
        return node

    def visit_SymbolRef(self, node):

        if node.type is not None:
            self.__add_entry(node.name, node.type)
        return node

    def visit_FunctionCall(self, node):
        if self._has_key(node.func):
            node.type = self._lookup(node.func)
        node.args = [self.visit(arg) for arg in node.args]
        return node

    def visit_BinaryOp(self, node):
        if isinstance(node.op, Op.Assign):
            node.left = self.visit(node.left)
            if isinstance(node.left, BinaryOp):
                return node
            node.right = self.visit(node.right)
            name = node.left
            value = node.right
            if hasattr(name, 'type') and name.type is not None:
                return node
            if hasattr(name, 'name') and not self._has_key(name.name):
                # temporary variable types can be derived from the variables
                # that they represent
                if name.name.startswith('____temp__'):
                    stripped_name = name.name.lstrip('____temp__')
                    if self._has_key(stripped_name):
                        node.left.type = self._lookup(stripped_name)
                    elif hasattr(value, 'get_type'):
                        node.left.type = value.get_type(self)
                elif hasattr(value, 'get_type'):
                    node.left.type = value.get_type()
                elif isinstance(value, String):
                    node.left.type = c_char_p()
                elif isinstance(value, SymbolRef):
                    node.left.type = self._lookup(value.name)
                elif isinstance(value, FunctionCall):
                    if self._has_key(value.func):
                        node.left.type = self._lookup(value.func)

                self.__add_entry(node.left.name, node.left.type)
        return node



import tempfile

__author__ = 'Chick Markley'


class DotManager(object):
    """
    take ast and return an ipython image file
    """

    @staticmethod
    def dot_ast_to_image(ast_node):
        dot_text = ast_node.to_dot()
        return DotManager.dot_text_to_image(dot_text)

    @staticmethod
    def dot_ast_to_browser(ast_node, file_name=None):
        dot_text = ast_node.to_dot()
        dot_output = DotManager.run_dot(dot_text)

        if file_name is None:
            with tempfile.NamedTemporaryFile(mode='wb', suffix=".png", delete=False) as f:
                f.write(dot_output)
                file_name = f.name
        else:
            with open(file_name, "wb") as f:
                f.write(dot_output)

        import subprocess
        subprocess.check_output(["open", file_name])

    @staticmethod
    def dot_ast_to_file(ast_node, file_name):
        dot_text = ast_node.to_dot()
        dot_output = DotManager.run_dot(dot_text)

        with open(file_name, "wb") as f:
            f.write(dot_output)

    @staticmethod
    def dot_text_to_image(text):
        try:
            from IPython.display import Image

            dot_output = DotManager.run_dot(text)
            return Image(dot_output, embed=True)
        except:
            return None

    @staticmethod
    def run_dot(code, options=None, output_format='png', file_name=None):
        # mostly copied from sphinx.ext.graphviz.render_dot
        import os
        from subprocess import Popen, PIPE
        try:
            from sphinx.util.osutil import EPIPE, EINVAL
        except ImportError:
            EPIPE, EINVAL = None, None
            raise RuntimeError(
                """
                graphical display of ASTs requires that the sphinx
                package be installed.
                """
            )

        if not options:
            options = []
        dot_args = ['dot'] + options + ['-T', output_format]
        if file_name:
            dot_args += ['>',file_name]

        p = None
        try:
            if os.name == 'nt':
                # Avoid opening shell window.
                # * https://github.com/tkf/ipython-hierarchymagic/issues/1
                # * http://stackoverflow.com/a/2935727/727827
                p = Popen(dot_args, stdout=PIPE, stdin=PIPE, stderr=PIPE,
                          creationflags=0x08000000)
            else:
                p = Popen(dot_args, stdout=PIPE, stdin=PIPE, stderr=PIPE)
        except OSError:
            raise RuntimeError(
                """
                Attempting to generate AST image. most likely dot (available through graphviz) is not installed
                or is not in your path
                """
            )
        went_wrong = False
        stdout, stderr = None, None  # for PEP-8
        try:
            # Graphviz may close standard input when an error occurs,
            # resulting in a broken pipe on communicate()
            stdout, stderr = p.communicate(code.encode('utf-8'))
        except (OSError, IOError) as err:
            if err.errno != EPIPE:
                raise
            went_wrong = True
        except IOError as err:
            if err.errno != EINVAL:
                raise
            went_wrong = True
        if went_wrong:
            # in this case, read the standard output and standard error streams
            # directly, to get the error message(s)
            stdout, stderr = p.stdout.read(), p.stderr.read()
            p.wait()
        if p.returncode != 0:
            raise RuntimeError('dot exited with error:\n[stderr]\n{0}'
                               .format(stderr.decode('utf-8')))
        return stdout



import logging

log = logging.getLogger(__name__)

from textwrap import dedent

import ctree
import time


def singleton(cls):
    instance = cls()
    instance.__call__ = lambda: instance
    return instance


def truncate(text):
    max_display_lines = ctree.CONFIG.getint("log", "max_lines_per_source")
    n_lines = len(text.splitlines())
    if n_lines <= max_display_lines:
        return text
    else:
        return dedent("""
        // %d lines suppressed, only %d allowed
        // consider adjusting config entry log.max_lines_per_source"""
                      % (n_lines, max_display_lines))


def lower_case_underscore_to_camel_case(string):
    """Convert string or unicode from lower-case underscore to camel-case"""
    # use string's class to work on the string to keep its type
    class_ = string.__class__
    return class_.join('', map(class_.capitalize, string.split('_')))


def flatten(obj):
    """Iterator for all objects arbitrarily nested in lists."""
    if isinstance(obj, (set, list)):
        for gen in map(flatten, obj):
            for elem in gen:
                yield elem
    elif isinstance(obj, (dict)):
        for gen in map(flatten, obj.itervalues()):
            for elem in gen:
                yield elem
    else:
        yield obj


def enumerate_flatten(obj_or_list):
    """Iterator for all objects arbitrarily nested in lists."""
    if isinstance(obj_or_list, list):
        for n, gen in enumerate(map(enumerate_flatten, obj_or_list)):
            for k, elem in gen:
                yield (n,)+k, elem
    else:
        yield (), obj_or_list


def highlight(code, language='c'):
    """Syntax-highlight code using pygments, if installed."""
    try:
        from pygments.formatters.terminal256 import Terminal256Formatter
        from pygments import highlight
    except ImportError:
        log.info("install pygments for syntax-highlighted output.")
        return code

    if language.lower() == 'llvm':
        from pygments.lexers.asm import LlvmLexer as TheLexer
    elif language.lower() == 'c':
        from pygments.lexers.compiled import CLexer as TheLexer
    elif language.lower() == 'diff':
        from pygments.lexers.text import DiffLexer as TheLexer
    elif language.lower() == 'ini':
        from pygments.lexers.text import IniLexer as TheLexer
    else:
        raise ValueError("Unrecognized highlight language: %s" % language)

    style = ctree.CONFIG.get('log', 'pygments_style')
    return highlight(code, TheLexer(), Terminal256Formatter(style=style))


class Timer:  # pragma: no cover
    def __enter__(self):
        self.start = time.clock()
        return self

    def __exit__(self, *args):
        self.interval = time.clock() - self.start



import unittest

from ctree.c.nodes import Constant, String


class TestConstants(unittest.TestCase):
    """Check that all constants convert properly."""

    def test_float_00(self):
        assert str(Constant(0)) == "0"

    def test_float_01(self):
        assert str(Constant(1)) == "1"

    def test_float_02(self):
        assert str(Constant(1.2)) == "1.2"

    def test_int_00(self):
        assert str(Constant(0)) == "0"

    def test_int_01(self):
        assert str(Constant(1)) == "1"

    def test_int_02(self):
        assert str(Constant(12)) == "12"

    def test_char_00(self):
        assert str(Constant("a")) == "'a'"

    def test_char_01(self):
        assert str(Constant("A")) == "'A'"

    def test_char_02(self):
        assert str(Constant("!")) == "'!'"


class TestStrings(unittest.TestCase):
    """Check that strings work."""

    def test_string_full(self):
        self.assertEqual(str(String("foo")), '"foo"')

    def test_string_empty(self):
        self.assertEqual(str(String("")), '""')

    def test_string_newline(self):
        self.assertEqual(str(String(r"\n")), r'"\n"')

    def test_string_tab(self):
        self.assertEqual(str(String(r"\t")), r'"\t"')

    def test_string_multi_two(self):
        self.assertEqual(str(String("foo", "bar")), '"foo" "bar"')

    def test_string_multi_three(self):
        self.assertEqual(str(String("foo", "bar", "baz")), '"foo" "bar" "baz"')

    def test_string_none(self):
        self.assertEqual(str(String()), '""')



import unittest

from itertools import islice

# class TestNullTuningDriver(unittest.TestCase):
#     def test_import(self):
#         import ctree.tune

#     def test_null_driver_stream(self):
#         from ctree.tune import NullTuningDriver

#         driver = NullTuningDriver()
#         for cfg in islice(driver.configs, 4):
#             self.assertDictEqual(cfg, {})

#     def test_null_driver_report(self):
#         from ctree.tune import NullTuningDriver

#         driver = NullTuningDriver()
#         for cfg in islice(driver.configs, 4):
#             driver.report(time=0.4)


try:
    import opentuner
except ImportError:
    HAVE_OPENTUNER = False
else:
    HAVE_OPENTUNER = True


class TestOpenTunerDriver(unittest.TestCase):
    @unittest.skipUnless(HAVE_OPENTUNER, "OpenTuner not in PYTHONPATH.")
    def test_import(self):
        from ctree.opentuner.driver import OpenTunerDriver

    @unittest.skipUnless(HAVE_OPENTUNER, "OpenTuner not in PYTHONPATH.")
    def test_1d_config(self):
        from ctree.opentuner.driver import OpenTunerDriver
        from opentuner.search.objective import MinimizeTime
        from opentuner.search.manipulator import (
            ConfigurationManipulator,
            IntegerParameter,
        )

        cfg = ConfigurationManipulator()
        cfg.add_parameter( IntegerParameter("foo", 0, 199) )
        driver = OpenTunerDriver(manipulator=cfg, objective=MinimizeTime())

        unsearched = set(range(200))
        for cfg in islice(driver.configs, 20):
            val = cfg["foo"]
            driver.report(time=val)
            unsearched.remove(val)
        self.assertEqual(len(unsearched), 180)

    # keep it to one test now because we're having problems
    # with multiple threads and the opentuner sqlite database.


class TestBruteForceTuningDriver(unittest.TestCase):
    def test_import(self):
        from ctree.tune import BruteForceTuningDriver

    def test_bruteforce_driver_1d(self):
        from ctree.tune import (
            BruteForceTuningDriver,
            IntegerParameter,
            MinimizeTime,
        )

        objective = MinimizeTime()
        params = [IntegerParameter("foo", 0, 100)]
        driver = BruteForceTuningDriver(params, objective)

        unsearched = set(range(0, 100))
        for config in islice(driver.configs, 100):
            unsearched.remove(config["foo"])
            driver.report(time=0.4)
        self.assertSetEqual(unsearched, set())

    def test_bruteforce_driver_2d(self):
        from ctree.tune import (
            BruteForceTuningDriver,
            IntegerParameter,
            MinimizeTime,
        )

        params = [
            IntegerParameter("foo", 0, 10),
            IntegerParameter("bar", 0, 10),
        ]
        driver = BruteForceTuningDriver(params, MinimizeTime())

        unsearched = set(range(0, 100))
        for config in islice(driver.configs, 100):
            entry = config["foo"] * 10 + config["bar"]
            unsearched.remove(entry)
            driver.report(time=0.4)
        self.assertSetEqual(unsearched, set())

    def test_bruteforce_driver_2d_parabola(self):
        from ctree.tune import (
            BruteForceTuningDriver,
            IntegerParameter,
            MinimizeTime,
        )

        params = [
            IntegerParameter("x", 0, 10),
            IntegerParameter("y", 0, 10),
        ]
        driver = BruteForceTuningDriver(params, MinimizeTime())

        for config in islice(driver.configs, 100):
            # report height on inverted paraboloid with global min at (3,4)
            x, y = config["x"], config["y"]
            z = (x-3)**2 + (y-4)**2 + 1
            driver.report(time=z)

        for config in islice(driver.configs, 10):
            self.assertEqual((config["x"], config["y"]), (3, 4))

    def test_bruteforce_driver_other_params(self):
        from ctree.tune import (
            BruteForceTuningDriver,
            IntegerParameter,
            BooleanParameter,
            EnumParameter,
            MinimizeTime,
        )

        params = [
            IntegerParameter("foo", 0, 10),
            BooleanParameter("bar"),
            EnumParameter("baz", ['monty', 'python', 'rocks']),
        ]
        driver = BruteForceTuningDriver(params, MinimizeTime())

        nConfigs = 10*2*3
        configs = list(islice(driver.configs, nConfigs))
        self.assertEqual(len(configs), nConfigs)



from ctypes import *

from util import CtreeTest
from ctree.c.nodes import *


class TestCastOps(CtreeTest):
    def setUp(self):
        self.foo = SymbolRef('foo')

    def test_void(self):
        tree = Cast(c_void_p(), self.foo)
        self._check_code(tree, "(void*) foo")

    def test_int(self):
        tree = Cast(c_long(), self.foo)
        self._check_code(tree, "(long) foo")

    def test_int_p(self):
        tree = Cast(POINTER(c_long)(), self.foo)
        self._check_code(tree, "(long*) foo")



__author__ = 'chick'

import os
import stat
import inspect
import shutil

from mako.template import Template


class Builder:
    """
    Class that creates a directory and file hierarchy based on a template directory
    ordinary files are copied as is
    *.mako files are rendered with mako into files with the .mako removed
    if the template directory name is code then it is changed to the target_base
    """
    verbose_key = 'verbose'

    def __init__(self, template_family, target_base, **kwargs):
        self.target_base = target_base
        self.template_family = template_family
        self.verbose = False

        if kwargs[Builder.verbose_key]:
            self.verbose = kwargs[Builder.verbose_key]

    def build(self, template_dir, target_dir, depth=0):
        """
        walks the template_dir
        each directory founds is created in associated target_dir
        each *.mako file is processed as a template and created without .mako
        other files are copied as is
        """

        make_executable = False

        def indent_print(s):
            if self.verbose:
                print (' ' * depth) + s

        if template_dir is None:
            a = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
            b = 'templates'
            c = self.template_family

            template_dir = os.path.join(a, b, c) #
            #    os.path.abspath(inspect.getfile(inspect.currentframe())),
            #    'templates',
            #    self.command_name,
            #)
            target_dir = self.target_base

        indent_print("template dir is %s" % template_dir)
        indent_print("target dir is %s" % target_dir)

        try:
            os.makedirs(target_dir)
        except OSError as exception:
            print ("Unable to create %s error (%d) %s" % \
                (target_dir,exception.errno,exception.strerror))
            exit(1)

        if target_dir[-4:] == '/bin':
            make_executable = True

        files = os.listdir(template_dir)
        indent_print("files " + ",".join(files))
        for file in files:
            source_file = os.path.join(template_dir, file)
            target_file = os.path.join(target_dir, file)

            if os.path.isfile(source_file):
                if source_file.endswith('.mako'):
                    template = Template(filename=source_file)
                    file_name = target_file[:-5]

                    f1=open(file_name,'w+')
                    indent_print("Rendering %s" % file_name)
                    print >>f1, template.render(specializer_name=self.target_base)
                    f1.close()
                else:
                    indent_print("processing ordinary file %s" % source_file)
                    if file != '.gitignore':
                        shutil.copyfile(source_file, target_file)
                        st = os.stat(target_file)
                        os.chmod(target_file, st.st_mode | stat.S_IEXEC | stat.S_IXOTH | stat.S_IXGRP)
            elif os.path.isdir(source_file):
                indent_print("processing directory %s" % file)
                destination = target_file if file != 'specializer_package' else os.path.join(target_dir, self.target_base)

                self.build(source_file, destination, depth+1)



"""
Macros for using OpenMP.
"""

from ctree.c.nodes import FunctionCall, SymbolRef, Block
from ctree.cpp.nodes import CppInclude
from ctree.omp.nodes import OmpParallelSections, OmpSection

def omp_get_num_threads():
    return FunctionCall(SymbolRef("omp_get_num_threads"), [])

def omp_get_thread_num():
    return FunctionCall(SymbolRef("omp_get_thread_num"), [])

def omp_get_wtime():
    return FunctionCall(SymbolRef("omp_get_wtime"), [])

def IncludeOmpHeader():
    return CppInclude("omp.h")


def parallelize_tasks(dag):
    """
    Returns an AST that computes the entries in dag in parallel using
    omp sections. Dag must consist of:
    1) tuples, implying elements must be executed sequentially,
    2) frozensets, implying elements can be executed in parallel,
    3) ASTs, the contents themselves.
    """
    if isinstance(dag, tuple):
        sched = []
        for node in dag:
            sched.extend(parallelize_tasks(node))
        return sched
    elif isinstance(dag, frozenset):
        sched = [OmpSection(body=parallelize_tasks(node)) for node in dag]
        return [OmpParallelSections(sections=sched)]
    else:
        return [dag]




import os
import unittest
from textwrap import dedent

from ctree.templates.nodes import StringTemplate, FileTemplate
from ctree.c.nodes import Constant, While
import fixtures


class TestStringTemplates(unittest.TestCase):
    def _check(self, tree, expected):
        actual = tree.codegen()
        self.assertEqual(actual, dedent(expected))

    def test_no_template_args(self):
        tree = StringTemplate("empty", {})
        self._check(tree, "empty")

    def test_simple_template_one(self):
        tree = StringTemplate("return $one", {'one': Constant(1)})
        self._check(tree, "return 1")

    def test_simple_template_two(self):
        tree = StringTemplate("return $one $two", {
            'one': Constant(1),
            'two': Constant(2),
        })
        self._check(tree, "return 1 2")

    def test_dotgen(self):
        tree = StringTemplate("return $one $two", {
            'one': Constant(1),
            'two': Constant(2),
        })
        dot = tree.to_dot()

    def test_indent_0(self):
        d = {'cond': Constant(1)}
        t = """\
        while($cond)
            printf("hello");
        """
        tree = While(Constant(0), [StringTemplate(t, d)])
        self._check(tree, """\
        while (0) {
            while(1)
                printf("hello");
        }""")

    def test_indent_1(self):
        d = {'cond': Constant(1)}
        t = """\
        while($cond)
            printf("hello");
        """
        tree = StringTemplate(t, d)
        self._check(tree, """\
        while(1)
            printf("hello");""")

    def test_child_single_list(self):
        d = {
            'stmts': [Constant(1)]
        }
        t = """$stmts"""
        tree = StringTemplate(t, d)
        self._check(tree, "1")


    def test_child_flat_list(self):
        d = {
            'stmts': [Constant(1), Constant(2)]
        }
        t = """$stmts"""
        tree = StringTemplate(t, d)
        self._check(tree, """\
        1;
        2;""")

    def test_child_nested_list(self):
        d = {
            'stmts': [[Constant(1)], Constant(2)]
        }
        t = """$stmts"""
        tree = StringTemplate(t, d)
        self._check(tree, """\
        1;
        2;""")

    def test_template_with_transformer(self):
        from ctree.visitors import NodeTransformer
        from ctree.c.nodes import String, SymbolRef

        template = "char *str = $val"
        template_args = {
            'val': SymbolRef("hello"),
        }
        tree = StringTemplate(template, template_args)
        self._check(tree, 'char *str = hello')

        class SymbolsToStrings(NodeTransformer):
            def visit_SymbolRef(self, node):
                return String(node.name)

        tree = SymbolsToStrings().visit(tree)
        self._check(tree, 'char *str = "hello"')

    def test_template_parent_pointers(self):
        from ctree.c.nodes import SymbolRef

        symbol = SymbolRef("hello")
        template = "char *str = $val"
        template_args = {
            'val': symbol,
        }
        node = StringTemplate(template, template_args)
        self.assertIs(symbol.parent, node)

    def test_template_parent_pointers_with_transformer(self):
        from ctree.visitors import NodeTransformer
        from ctree.c.nodes import String, SymbolRef

        template = "char *str = $val"
        template_args = {
            'val': SymbolRef("hello"),
        }

        class SymbolsToStrings(NodeTransformer):
            def visit_SymbolRef(self, node):
                return String(node.name)

        tree = StringTemplate(template, template_args)
        tree = SymbolsToStrings().visit(tree)

        template_node, string = tree, tree.val
        self.assertIs(string.parent, template_node)


class TestFileTemplates(unittest.TestCase):
    def _check(self, tree, expected):
        actual = tree.codegen()
        self.assertEqual(actual, dedent(expected))

    def test_simple_file_template(self):
        from ctree.c.nodes import String
        path = os.path.join(*(fixtures.__path__ + ["templates", "printf.tmpl.c"]))
        tree = FileTemplate(path, {'fmt': String('Hello, world!')})
        self._check(tree, 'printf("Hello, world!");')

    def test_file_template_dotgen(self):
        from ctree.c.nodes import String
        path = os.path.join(*(fixtures.__path__ + ["templates", "printf.tmpl.c"]))
        tree = FileTemplate(path, {'fmt': String('Hello, world!')})
        tree.to_dot()



import unittest

from ctree.c.nodes import *


class TestReturns(unittest.TestCase):
    def _check(self, tree, expected):
        actual = str(tree)
        self.assertEqual(actual, expected)

    def test_return_void(self):
        tree = Return()
        self._check(tree, "return")

    def test_return_int(self):
        tree = Return(Constant(1))
        self._check(tree, "return 1")

    def test_return_sym(self):
        tree = Return(SymbolRef('foo'))
        self._check(tree, "return foo")


class TestIfs(unittest.TestCase):
    def setUp(self):
        self.cond = Constant(1)
        self.then = [Assign(SymbolRef('foo'), SymbolRef('bar'))]
        self.elze = [Assign(SymbolRef('spam'), SymbolRef('eggs'))]

    def _check(self, tree, expected):
        actual = str(tree)
        self.assertEqual(actual, expected)

    def test_ifthen(self):
        tree = If(self.cond, self.then)
        self._check(tree, """if (1) {
    foo = bar;
}""")

    def test_ifthenelse(self):
        tree = If(self.cond, self.then, self.elze)
        self._check(tree, """if (1) {
    foo = bar;
} else {
    spam = eggs;
}""")


class TestWhiles(unittest.TestCase):
    def setUp(self):
        self.cond = Constant(2)
        self.body = [SymbolRef('foo')]

    def _check(self, tree, expected):
        actual = str(tree)
        self.assertEqual(actual, expected)

    def test_while(self):
        tree = While(self.cond, self.body)
        self._check(tree, """while (2) {
    foo;
}""")

    def test_dowhile(self):
        tree = DoWhile(self.body, self.cond)
        self._check(tree, """do {
    foo;
} while (2)""")


class TestFor(unittest.TestCase):
    def _check(self, tree, expected):
        actual = str(tree)
        self.assertEqual(actual, expected)

    def test_for_00(self):
        init = Assign(SymbolRef("foo"), Constant(0))
        test = Lt(SymbolRef("foo"), Constant(10))
        incr = PostInc(SymbolRef("foo"))
        body = [FunctionCall(SymbolRef("printf"), [SymbolRef("foo")])]
        tree = For(init, test, incr, body)
        self._check(tree, """for (foo = 0; foo < 10; foo ++) {
    printf(foo);
}""")



"""
DOT generator for C constructs.
"""

from ctree.dotgen import DotGenLabeller
from ctree.types import codegen_type


class CDotGenLabeller(DotGenLabeller):
    """
    Manages generation of DOT.
    """

    def visit_SymbolRef(self, node):
        s = r""
        if node._global:
            s += r"__global "
        if node._local:
            s += r"__local "
        if node._const:
            s += r"__const "
        if node.type is not None:
            s += r"%s " % codegen_type(node.type)
        s += r"%s" % node.name
        return s

    def visit_FunctionDecl(self, node):
        s = r""
        if node.static:
            s += r"static "
        if node.inline:
            s += r"inline "
        if node.kernel:
            s += r"__kernel "
        s += r"%s %s(...)" % (codegen_type(node.return_type), node.name)
        return s

    def visit_Constant(self, node):
        return str(node.value)

    def visit_String(self, node):
        return r'\" \"'.join(node.values)

    def visit_CFile(self, node):
        return node.get_filename()

    def visit_NdPointer(self, node):
        s = "dtype: %s\n" % node.ptr.dtype
        s += "ndim, shape:    %s, %s\n" % (node.ptr.ndim, node.ptr.shape)
        s += "flags: %s" % node.ptr.flags
        return s

    def visit_BinaryOp(self, node):
        return type(node.op).__name__

    def visit_UnaryOp(self, node):
        return type(node.op).__name__



"""
Make sure the examples work.

The 'examples' directory needs to be in PYTHONPATH for these to pass.
"""

import unittest

try:
  import examples
except ImportError:
  HAVE_EXAMPLES = False
else:
  HAVE_EXAMPLES = True

@unittest.skipUnless(HAVE_EXAMPLES, "$CTREE/examples not in PYTHONPATH")
class TestAllExamples(unittest.TestCase):
    def test_AstToDot(self):
        from examples import AstToDot
        AstToDot.main()

    def test_Fibonacci(self):
        from examples import Fibonacci
        Fibonacci.main()

    def test_ArrayDoubler(self):
        from examples import ArrayDoubler
        ArrayDoubler.main()

    def test_TemplateDoubler(self):
        from examples import TemplateDoubler
        TemplateDoubler.main()

    def test_SimpleTranslator(self):
        from examples import SimpleTranslator
        SimpleTranslator.main()

    @unittest.skip("suppressed opentuner test")
    def test_TuningSpecializer(self):
        from examples import TuningSpecializer
        TuningSpecializer.main()

    def test_OclDoubler(self):
        from examples import OclDoubler
        OclDoubler.main()

    @unittest.skip("under development")
    def test_StencilGrid(self):
        from examples import StencilGrid
        StencilGrid.main()



"""
DOT generation for template nodes.
"""

import os

from ctree.dotgen import DotGenLabeller


class TemplateDotLabeller(DotGenLabeller):
    """
    Visitor to generator DOT.
    """
    def visit_StringTemplate(self, node):
        return "template: <<<\n%s\n>>>" % \
            node._template.template.replace("\n", "\\n").replace('"', r"\"")

    def visit_FileTemplate(self, node):
        return os.path.basename(node._template_path)



"""
Support for Python ast nodes in the 'ast' module.
"""

import ast

# ---------------------------------------------------------------------------
# dot generator

from ctree.dotgen import DotGenLabeller


class PyDotLabeller(DotGenLabeller):  # pragma: no cover
    """
    Manages generation of DOT.
    """

    def visit_arg(self, node):
        s = "name: %s" % node.arg
        if node.annotation and not isinstance(node.annotation, ast.AST):
            s += "\nannotation: %s" % self._qualified_name(node.annotation)
        return s

    def visit_FunctionDef(self, node):
        return "name: %s" % node.name

    def visit_Num(self, node):
        return "n: %s" % node.n

    def visit_Name(self, node):
        return "id: %s" % node.id

    def visit_Attribute(self, node):
        return "attr: %s" % node.attr

    def visit_Str(self, node):
        return "str: %s" % node.s



"""
Code generation for SIMD.
"""

from ctree.codegen import CodeGenVisitor


class SimdCodeGen(CodeGenVisitor):
    """
    Visitor to generate SIMD code.
    """
    def visit_m256d(self, node):
        return "__m256d"



"""
DOT generation for C preprocessor directives.
"""

from ctree.dotgen import DotGenLabeller


class CppDotLabeller(DotGenLabeller):
    """
    Visitor to generator DOT.
    """

    def visit_CppInclude(self, node):
        if node.angled_brackets:
            return "target: <%s>" % node.target
        else:
            return 'target: "%s"' % node.target

    def visit_CppComment(self, node):
        return "// " + node.text.replace('"', r"\"")



"""
Cilk nodes supported by ctree.
"""

from ctree.nodes import CtreeNode


class CilkNode(CtreeNode):
    """Base class for all Cilk nodes supported by ctree."""

    def codegen(self, indent=0):
        from ctree.cilk.codegen import CilkCodeGen

        return CilkCodeGen(indent).visit(self)

    def _to_dot(self, _):
        from ctree.cilk.dotgen import CilkDotLabeller

        return CilkDotLabeller().visit(self)



"""
Defines the hierarchy of AST nodes.
"""

import logging
import os.path

log = logging.getLogger(__name__)

import ast
import collections

from ctree.codegen import CodeGenVisitor
from ctree.dotgen import DotGenVisitor, DotGenLabeller
from ctree.util import flatten
import ctree
import os


class CtreeNode(ast.AST):
    """Base class for all AST nodes in ctree."""
    _fields = []

    def __init__(self):
        """Initialize a new AST Node."""
        super(CtreeNode, self).__init__()
        self.deleted = False
        self._force_parentheses = False

    def __str__(self):
        return self.codegen()

    def codegen(self, indent=0):
        raise Exception("Node class %s should override codegen()" % type(self))

    def delete(self):
        self.codegen = self.no_code_gen
        self.deleted = True

    def no_code_gen(self, *args):
        return ""

    def to_dot(self):
        """Retrieve the AST in DOT format for vizualization."""
        return "digraph mytree {\n%s}" % self._to_dot()

    def _to_dot(self):
        """Retrieve the AST in DOT format for vizualization."""
        return DotGenVisitor().visit(self)

    def _requires_semicolon(self):
        """When coverted to a string, this node should be followed by a semicolon."""
        return True

    def find_all(self, node_class, **kwargs):
        """
        Returns a generator that yields all nodes of type
        'node_class' type whose attributes match those specified
        in kwargs. For example, all FunctionDecls with name 'fib'
        can be accessed via:
            my_ast.find_all(FunctionDecl, name="fib")
        """

        def pred(node):
            if isinstance(node, node_class):
                for attr, value in kwargs.items():
                    try:
                        if getattr(node, attr) != value:
                            break
                    except AttributeError:
                        break
                else:
                    return True
            return False

        return self.find_if(pred)

    def find(self, node_class, **kwargs):
        """
        Returns one node of type 'node_class' whose attributes
        match those specified in kwargs, or None if no nodes
        can be found.
        """
        matching = self.find_all(node_class, **kwargs)
        try:
            return next(matching)
        except StopIteration:
            return None

    def find_if(self, pred):
        """
        Returns all nodes satisfying the given predicate,
        or None if no satisfactory nodes are found. The search
        starts from the current node.
        """
        for node in ast.walk(self):
            if pred(node):
                yield node

    def lift(self, **kwargs):
        for key, val in kwargs.items():
            attr = "_lift_%s" % key
            setattr(self, attr, getattr(self, attr, []) + val)
            type(self)._fields.append(attr)

    def __eq__(self, other):
        """Two nodes are equal if their attributes are equal."""
        return self.__dict__ == getattr(other, '__dict__', None)


# ---------------------------------------------------------------------------
# Common nodes

class CommonNode(CtreeNode):
    """Miscellaneous IR nodes."""

    def codegen(self, indent=0):
        return CommonCodeGen(indent).visit(self)

    def label(self):
        return CommonDotGen().visit(self)


class Project(CommonNode):
    """Holds a list files."""
    _fields = ['files']

    def __init__(self, files=None, indent=0, compilation_dir = ''):
        self.files = files if files else []
        super(Project, self).__init__()
        self.compilation_dir = compilation_dir
        self.indent = indent

    def codegen(self, indent=0):
        """
        Code generates each file in the project and links their
        bytecode together to get the master bytecode file.
        """
        from ctree.jit import JitModule

        self._module = JitModule()

        # now that we have a concrete compilation dir, resolve references to it
        from ctree.transformations import ResolveGeneratedPathRefs
        #
        # resolver = ResolveGeneratedPathRefs(self._module.compilation_dir)
        # self.files = [resolver.visit(f) for f in self.files]
        # if resolver.count:
        #     log.info("automatically resolved %d GeneratedPathRef node(s).", resolver.count)

        # transform all files into llvm modules and link them into the master module
        for f in self.files:
            submodule = f._compile(f.codegen())
            if submodule:
                self._module._link_in(submodule)
        return self._module

    @property
    def module(self):
        if self._module:
            return self._module
        return self.codegen(indent=self.indent)


class File(CommonNode):
    """Holds a list of statements."""
    _fields = ['body']
    _empty = None


    def __init__(self, name="generated", body=None, path = None):
        self.name = name
        self.body = body or []
        self.config_target = 'c'
        self.path = path or ctree.CONFIG.get('jit','COMPILE_PATH')
        self._program_hash = None

    @property
    def empty(self):
        if not self._empty:
            self._empty = type(self)(name=self.name, path=self.path).codegen()
        return self._empty


    @property
    def path(self):
        return self._path

    @path.setter
    def path(self, value):
        self._path = value
        if not os.path.exists(self._path):
            os.makedirs(self._path)

    def get_hash_filename(self):
        return "%s.%s.sha" % (self.name, self._ext)


    @property
    def program_hash(self):
        if not os.path.exists(os.path.join(self.path, self.get_hash_filename())):
            return False
        if self._program_hash:
            return self._program_hash
        with open(os.path.join(self.path, self.get_hash_filename())) as h_file:
            return h_file.read().strip()

    @program_hash.setter
    def program_hash(self, value):
        self._program_hash = value
        with open(os.path.join(self.path, self.get_hash_filename()), 'w') as h_file:
            h_file.write(value)

    def codegen(self, *args):
        """Convert this substree into program text (a string)."""
        raise Exception("%s should override codegen()." % type(self))

    def _compile(self, program_text, compilation_sub_dir):
        """Construct an LLVM module with the translated contents of this file."""
        raise Exception("%s should override _compile()." % type(self))

    def get_generated_path_ref(self):
        """Returns an object that can resolve the full file path at compile time."""
        return GeneratedPathRef(self)

    def get_filename(self):
        return "%s.%s" % (self.name, self._ext)


class GeneratedPathRef(CommonNode):
    """Represents a path to a generated file."""
    _force_parentheses = False

    def __init__(self, target_file=None):
        assert isinstance(target_file, File), \
            "Cannot create a GeneratedPathRef to a %s, must be a File." % repr(target_file)
        self.target = target_file


class CommonCodeGen(CodeGenVisitor):
    """Manages conversion of all common nodes to txt."""

    def visit_File(self, node):
        return ";\n".join(map(str, node.body)) + ";\n"

    def visit_GeneratedPathRef(self, node):
        return '"%s"'% (os.path.join(node.target.path, node.target.get_filename()))
        # raise Exception("Unresolved GeneratedPathRefs to file %s." % (node.target.get_filename()))

    def visit_Pass(self, node):
        return ""


class CommonDotGen(DotGenLabeller):
    """Manages coversion of all common nodes to dot."""

    def visit_GeneratedPathRef(self, node):
        return "target: %s" % node.target.get_filename()



import unittest
import ctypes as ct

from ctree.c.nodes import *
from ctree.precedence import *


class TestPrecedence(unittest.TestCase):
    def setUp(self):
        self.args = (SymbolRef('a'), SymbolRef('b'), SymbolRef('c'))

    def _check(self, tree, expected):
        actual = str(tree)
        self.assertEqual(actual, expected)

    def test_addmul_rhs(self):
        a, b, c = self.args
        tree = Add(a, Mul(b, c))
        self._check(tree, "a + b * c")

    def test_addmul_lhs(self):
        a, b, c = self.args
        tree = Add(Mul(a, b), c)
        self._check(tree, "a * b + c")

    def test_muladd_rhs(self):
        a, b, c = self.args
        tree = Mul(a, Add(b, c))
        self._check(tree, "a * (b + c)")

    def test_muladd_lhs(self):
        a, b, c = self.args
        tree = Mul(Add(a, b), c)
        self._check(tree, "(a + b) * c")

    def test_muladd_lhs_rhs(self):
        a, b, c = self.args
        tree = Mul(Add(a, b), Add(c, a))
        self._check(tree, "(a + b) * (c + a)")

    def test_muladd_nested(self):
        a, b, c = self.args
        tree = Add(a, Mul(b, Add(c, a)))
        self._check(tree, "a + b * (c + a)")

    def test_preinc_muladd(self):
        a, b, c = self.args
        tree = Mul(PreInc(Add(a, b)), c)
        self._check(tree, "++ (a + b) * c")

    def test_not_and_or(self):
        a, b, c = self.args
        tree = Not(Or(And(a, b), c))
        self._check(tree, "! (a && b || c)")

    def test_sub_postinc(self):
        a, b, c = self.args
        tree = Sub(Sub(PostInc(a)), b)
        self._check(tree, "- a ++ - b")

    def test_sub_unary_binary(self):
        a, b, c = self.args
        tree = Sub(Sub(a, b))
        self._check(tree, "- (a - b)")

    def test_sub_add_unary_binary(self):
        a, b, c = self.args
        tree = Add(Sub(Sub(a, Add(b))), c)
        self._check(tree, "- (a - + b) + c")

    def test_postinc_unary(self):
        a, b, c = self.args
        tree = PostInc(Sub(a))
        self._check(tree, "(- a) ++")

    def test_cast1(self):
        a, b, c = self.args
        tree = Add(Cast(ct.c_int(), a), b)
        self._check(tree, "(int) a + b")

    def test_cast2(self):
        a, b, c = self.args
        tree = Cast(ct.c_int(), Add(a, b))
        self._check(tree, "(int) (a + b)")


class TestAssociativityPrecedence(unittest.TestCase):
    """
    Tests insertion of parentheses for trees of operators that
    have the same precedence, and therefore we need to consider
    their associativity.
    """

    def setUp(self):
        self.args = (SymbolRef('a'), SymbolRef('b'), SymbolRef('c'))

    def _check(self, tree, expected):
        actual = str(tree)
        self.assertEqual(actual, expected)

    def test_subadd_rhs(self):
        a, b, c = self.args
        tree = Sub(a, Add(b, c))
        self._check(tree, "a - (b + c)")

    def test_subadd_lhs(self):
        a, b, c = self.args
        tree = Sub(Add(a, b), c)
        self._check(tree, "a + b - c")

    def test_complicated_00(self):
        a, b, c = self.args
        tree = Sub(Add(a, Sub(b, a)), Add(Sub(a, b), c))
        self._check(tree, "a + (b - a) - (a - b + c)")

    def test_augassign(self):
        a, b, c = self.args
        tree = AddAssign(a, AddAssign(b, c))
        self._check(tree, "a += b += c")

    def test_ternary_cond(self):
        a, b, c = self.args
        cond = TernaryOp(a, b, c)
        tree = TernaryOp(cond, c, b)
        self._check(tree, "(a ? b : c) ? c : b")

    def test_ternary_then(self):
        a, b, c = self.args
        then = TernaryOp(a, b, c)
        tree = TernaryOp(c, then, a)
        self._check(tree, "c ? (a ? b : c) : a")

    def test_ternary_else(self):
        a, b, c = self.args
        elze = TernaryOp(a, b, c)
        tree = TernaryOp(c, b, elze)
        self._check(tree, "c ? b : a ? b : c")

    def test_bad_precedence_arg(self):
        with self.assertRaises(Exception):
            get_precedence(Constant(2.3))

    def test_bad_op(self):
        a, b, c = self.args
        tree = BinaryOp(a, b, c)
        with self.assertRaises(Exception):
            get_precedence(tree)

    def test_bad_associativity_arg(self):
        with self.assertRaises(Exception):
            is_left_associative(Constant(2.3))



import unittest

from ctree.c.nodes import *


class TestAugAssigns(unittest.TestCase):
    def setUp(self):
        self.foo, self.bar = SymbolRef("foo"), SymbolRef("bar")

    def _check(self, node_cls, expected_string):
        node = node_cls(self.foo, self.bar)
        self.assertEqual(str(node), expected_string)

    def test_add(self):
        self._check(AddAssign, "foo += bar")

    def test_sub(self):
        self._check(SubAssign, "foo -= bar")

    def test_mul(self):
        self._check(MulAssign, "foo *= bar")

    def test_div(self):
        self._check(DivAssign, "foo /= bar")

    def test_mod(self):
        self._check(ModAssign, "foo %= bar")

    def test_bitxor(self):
        self._check(BitXorAssign, "foo ^= bar")

    def test_bitand(self):
        self._check(BitAndAssign, "foo &= bar")

    def test_bitor(self):
        self._check(BitOrAssign, "foo |= bar")

    def test_bitshl(self):
        self._check(BitShLAssign, "foo <<= bar")

    def test_bitshr(self):
        self._check(BitShRAssign, "foo >>= bar")



import unittest

from ctree.util import truncate
from ctree.util import lower_case_underscore_to_camel_case
from ctree.util import singleton


class TestTruncate(unittest.TestCase):
    def test_no_truncate(self):
        text = 'arst'
        self.assertEqual(truncate(text), text)

    def test_trunctate(self):
        text = 'a\na\na\na\na\na\na\na\na\na\na\na \
                \na\na\na\na\na\na\na\na\na\na\na \
                \na\na\na\na\na\na\na\na\na\na'
        self.assertNotEqual(truncate(text), text)


@singleton
class Single(object):
    def id(self):
        return self.__hash__()


class TestUtil(unittest.TestCase):
    def _check(self, actual, expected):
        self.assertEqual(actual, expected)

    def test_singleton(self):
        s1 = Single
        s2 = Single

        self.assertRaises(TypeError, Single)

        self.assertEqual(s1.id, s2.id)

    def test_truncate(self):
        s = "foo"
        s_truncated = truncate(s)
        self.assertEqual(s, s_truncated)

        many_lines = "\n".join(map(lambda x: "line %d" % x, [l for l in range(500)]))
        less_lines = truncate(many_lines)

        self.assertTrue("lines suppressed" in less_lines)

    def test_lower_case_to_camel_case(self):
        s = "dog_cat"
        cc = lower_case_underscore_to_camel_case(s)

        self.assertEqual(cc,"DogCat")


class TestLowerCaseUnderscoreToCamelCase(unittest.TestCase):
    def test_simple(self):
        text = 'this_is_a_name'
        self.assertEqual(
            lower_case_underscore_to_camel_case(text),
            'ThisIsAName'
        )




"""
DOT labeller for OpenMP.
"""

from ctree.dotgen import DotGenLabeller

# ---------------------------------------------------------------------------
# DOT generator

class OmpDotLabeller(DotGenLabeller):
    """
    Visitor to generator DOT.
    """
    pass



from ctree.nodes import CtreeNode
from ctree.util import flatten

class TemplateNode(CtreeNode):
    """Base class for all template nodes."""
    def __init__(self, template_txt, child_dict):
        """
        Create a new template node.

        :param template_txt: The template as a string.
        :param child_dict: A mapping between template keys \
        and subtrees (CtreeNodes).
        """
        from textwrap import dedent
        from string import Template
        from collections import namedtuple

        dedented_txt = dedent(template_txt)
        self._template = Template(dedented_txt)
        self._children = child_dict
        self._fields = child_dict.keys()
        super(TemplateNode, self).__init__()

    def codegen(self, indent=0):
        from ctree.templates.codegen import TemplateCodeGen

        return TemplateCodeGen(indent).visit(self)

    def label(self):
        from ctree.templates.dotgen import TemplateDotLabeller

        return TemplateDotLabeller().visit(self)

    def _requires_semicolon(self):
        return False

    def __setattr__(self, name, val):
        from ctree.nodes import CtreeNode

        if name == "_children":
            # set parent pointers in child_dict
            assert isinstance(val, dict)
            super(TemplateNode, self).__setattr__(name, val)
            for name, value in val.items():
                for child in flatten(value):
                    child.parent = self

        elif hasattr(self, "_children") and name in self._children:
            # insert into _children dictionary and set parent pointers
            self._children[name] = val
            if isinstance(val, CtreeNode):
                val.parent = self
        else:
            # do standard attribute resolution
            super(TemplateNode, self).__setattr__(name, val)

    def __getattr__(self, name):
        if name != "_children" and name in self._children:
            child = self._children[name]
            assert child.parent == self, "Encountered bad parent pointer to %s." % repr(child.parent)
            return self._children[name]
        raise AttributeError("'%s' has no attribute '%s'" % (type(self).__name__, name))

class StringTemplate(TemplateNode):
    """
    A template node that wraps Python's string.Template.
    """
    def __init__(self, template_txt="", child_dict={}):
        """
        Create a new template node.

        :param template_txt: The template as a string.
        :param child_dict: A mapping between template keys \
        and subtrees (CtreeNodes).
        """
        super(StringTemplate, self).__init__(template_txt, child_dict)

class FileTemplate(TemplateNode):
    """
    A template node that wraps Python's string.Template.
    """
    def __init__(self, template_path=".", child_dict={}):
        """
        Create a new template node.

        :param template_path: The path to the template file.
        :param child_dict: A mapping between template keys \
        and subtrees (CtreeNodes).
        """
        self._template_path = template_path
        with open(template_path, "r") as template_file:
            contents = template_file.read()
        super(FileTemplate, self).__init__(contents, child_dict)



import ctree.c.nodes as C
import ast


op_map = {
    C.Op.Add: lambda x, y: x + y,
    C.Op.Div: lambda x, y: x / y,
    C.Op.Mul: lambda x, y: x * y,
    C.Op.Lt: lambda x, y: x < y,
    C.Op.Sub: lambda x, y: x - y,
}


class ConstantFold(ast.NodeTransformer):
    """ TODO: Support all folding situations """
    def fold_add(self, node):
        if isinstance(node.left, C.Constant) and node.left.value == 0:
            return node.right
        elif isinstance(node.right, C.Constant) and node.right.value == 0:
            return node.left
        return node

    def fold_sub(self, node):
        if isinstance(node.left, C.Constant) and node.left.value == 0:
            return C.Sub(node.right)
        elif isinstance(node.right, C.Constant) and node.right.value == 0:
            return node.left
        return node

    def fold_mul(self, node):
        if isinstance(node.left, C.Constant) and node.left.value == 1:
            return node.right
        elif isinstance(node.right, C.Constant) and node.right.value == 1:
            return node.left
        elif isinstance(node.left, C.Constant) and node.left.value == 0:
            return node.left
        elif isinstance(node.right, C.Constant) and node.right.value == 0:
            return node.right
        return node

    def visit_BinaryOp(self, node):
        node.left = self.visit(node.left)
        node.right = self.visit(node.right)
        if isinstance(node.left, C.Constant) and \
                isinstance(node.right, C.Constant):
            return C.Constant(op_map[node.op.__class__](
                node.left.value, node.right.value))
        elif isinstance(node.op, C.Op.Add):
            return self.fold_add(node)
        elif isinstance(node.op, C.Op.Sub):
            return self.fold_sub(node)
        elif isinstance(node.op, C.Op.Mul):
            return self.fold_mul(node)
        return node



"""
DOT labeller for Cilk.
"""

from ctree.dotgen import DotGenLabeller

# ---------------------------------------------------------------------------
# DOT labeller

class CilkDotLabeller(DotGenLabeller):
    """
    Visitor to label DOT nodes.
    """
    pass



from ctree.types import CtreeType


class SimdType(CtreeType):
    """Base class for all SIMD Types."""

    def codegen(self, indent=0):
        from ctree.simd.codegen import SimdCodeGen

        return SimdCodeGen().visit(self)


class m256d(SimdType):
    pass



"""
Code generation for OpenCL.
"""

from ctree.codegen import CodeGenVisitor


class OclCodeGen(CodeGenVisitor):
    """
    Visitor to generate OpenCL code.
    """

    def visit_OclFile(self, node):
        stmts = self._genblock(node.body, insert_curly_brackets=False, increase_indent=False)
        return '// <file: %s>%s' % (node.get_filename(), stmts)



from copy import deepcopy

from util import CtreeTest
from fixtures.sample_asts import *
from ctree.transformations import Lifter

class TestLifter(CtreeTest):
    def test_nop(self):
        for py_fn, tree in SAMPLE_ASTS:
            transformed = Lifter().visit(deepcopy(tree))
            self._check_code(actual=tree, expected=transformed)

    def test_one_param(self):
        inner = SymbolRef("foo")
        inner.lift(params=[SymbolRef(inner.name, c_int())])

        tree = FunctionDecl(None, "fn", [], [
            Assign(inner, Constant(123)),
        ])

        tree = Lifter().visit(tree)

        self._check_code(actual=tree, expected="""\
        void fn(int foo) {
            foo = 123;
        }""")

    def test_two_params(self):
        inner0 = SymbolRef("foo")
        inner0.lift(params=[SymbolRef(inner0.name, c_int())])

        inner1 = SymbolRef("bar")
        inner1.lift(params=[SymbolRef(inner1.name, c_double())])

        tree = FunctionDecl(None, "fn", [], [
            Assign(inner0, Constant(123)),
            Assign(inner1, Constant(456.7)),
        ])

        tree = Lifter().visit(tree)

        self._check_code(actual=tree, expected="""\
        void fn(int foo, double bar) {
            foo = 123;
            bar = 456.7;
        }""")

    def test_one_include(self):
        tree = CFile("generated", [deepcopy(get_two_ast)])
        stmt = tree.find(FunctionDecl).defn[0]
        stmt.lift(includes=[CppInclude("stdio.h")])

        tree = Lifter().visit(tree)

        self._check_code(actual=tree, expected="""\
        // <file: generated.c>
        #include <stdio.h>
        long get_two() {
            return 2;
        };
        """)

    def test_multi_includes(self):
        tree = CFile("generated", [deepcopy(get_two_ast)])
        stmt0 = tree.find(FunctionDecl)
        stmt1 = stmt0.defn[0]

        stmt0.lift(includes=[CppInclude("stdio.h")])
        stmt1.lift(includes=[CppInclude("stdlib.h"), CppInclude("float.h")])

        tree = Lifter().visit(tree)

        self._check_code(actual=tree, expected="""\
        // <file: generated.c>
        #include <stdio.h>
        #include <stdlib.h>
        #include <float.h>
        long get_two() {
            return 2;
        };
        """)
